/*
***                           File main.c                         ***
***                        GeoFEST version 4.2                    ***
*** Copyright (c) 2002, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This file contains the program main entry point, the main
*** task driver, and modules for driving high-level functions
*** and interactions with the operator:
***
***          == main ==
***          == elastic ==
***          == time_step ==
***          == clear_stiff ==
***          == elgrp_loop ==
***          == node_load ==
***          == accumulate ==
***          == clr_smooth ==   - disabled -
***          == do_smooth ==   - disabled -
***          == completion ==
***          == wrt_save ==
***          == vrestart ==
***          == el_save ==
***          == check_monitor ==
***/

/*
***GeoFEST version 4.2***
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining 
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
$Log: main.c,v $
Revision 1.1.1.1  2002/07/12 19:05:14  jwp
Commiting GeoFEST to CVS at version 4.2

*/

#include <stdio.h>
#include "mydefs.h"
#define EXTERN
#include "finel.h"
#include "errcodes.h"
#include <math.h>

#define OK 0
#define KILL 1
#define SAVE_STOP 2
#define SAVE_GO 3
#define NCODES 4

   char  line[80] , fullname[80] ;
   int   step_count ;
   int   add_int() ;
   char  *mon_codes[NCODES] =
          { "OK" , "KILL" , "SAVE_STOP" , "SAVE_GO" } ;
       
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   main   ***************
*/

/*  INTERFACE:   */

     main(
          int argc ,      /* number of input arguments */
          char *argv[]    /* input argument strings */
         )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine main
** main prints a banner, 
** interprets up to two args as input and output files,
** scans two comment lines in the input, 
** and divides (and times) the processing
** into input, elastic solution, time-stepping, and output.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
       char CVS_Id[] = "@(#) CVS id $Id: main.c,v 1.1.1.1 2002/07/12 19:05:14 jwp Exp $, Revision: " REVNO;
       int  i ;
     
       printf(
       "\n\n\n************************************************\n\n") ;
       printf(
       "Welcome to the Geophysical Finite Element Simulation Tool (GeoFEST " REVNO ") (%s)\n"
               , argv[0] ) ;
 
       if( argc == 1 )
          {
           printf("What is the name of the input data file ?\n");
           scanf("%s",line);
           sprintf(fullname,"%s",line) ;
          }
       else
          {
           printf("Opening input file %s ...\n",argv[1]);
           sprintf(fullname,"%s",argv[1]) ;
          }

       if((in_file = fopen(fullname,"r")) == NULL)
          {
           printf("Input file couldn't be opened!\n") ;
           exit(1) ;
          }

       fscanf(in_file,"%s",line);
       printf("Opening output file %s ...\n",line);
       sprintf(fullname,"%s",line) ;
       if((out_file = fopen(fullname,"w")) == NULL)
          {
           printf("Output file couldn't be opened!\n") ;
           exit(1) ;
          }

/*  *** read and echo the pass-thru comment lines *** */

       fscanf(in_file,"%[^*]*\n",line) ;
       printf("%s\n",line) ;
       fprintf(out_file,"%s\n",line) ;

       fscanf(in_file,"%[^*]*\n",line) ;
       printf("%s\n",line) ;
       fprintf(out_file,"%s\n",line) ;


       input_phase() ;


       if( !time_data.restart )
          {
           elastic() ;
          }


       time_step() ;


       if(time_data.sav_state)
          wrt_save() ;

       fclose(in_file) ;
       fclose(out_file) ;
       printf("Normal program termination.\n") ;
       exit(0) ;
     } 
/****************************** end of main **********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   elastic   ***************
*/

/*  INTERFACE:   */

     elastic(
            )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine elastic
** elastic performs an elastic solution of the 
** finite element problem.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
    {
     printf("Starting elastic solution...\n") ;
     time_data.elastic = 1 ;
     clear_real(global.displ,fe_sys.numnp*fe_sys.ndof) ;
     clear_real(global.del_displ,fe_sys.numnp*fe_sys.ndof) ;
     clear_real(force.full_rhs,fe_sys.neq) ;
     clear_stiff() ;
     node_load(ZERO,YES) ;

     elgrp_loop( BC_ELAS ) ;
     time_data.do_slip = 1 ;
     elgrp_loop( FORMS_ELAS ) ;
     elgrp_loop( RHS_ELAS ) ;

     if (fe_sys.solver == DIRECT)
        solver( FACBACK ) ;
     else if (fe_sys.solver == PCG)
        solver( ITER ) ;
     accumulate() ;
     elgrp_loop( E_STRESS ) ;
     printf("Elastic solution complete.\n") ;
    }
/*************************** end of elastic **********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   time_step   ***************
*/

/*  INTERFACE:   */

     time_step(
              )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine time_step
** time_step computes the visco-elastic time stepping solutions
** to the finite element problem.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/
    {
      real  time , dt , alpha ;
      int  time_group , nsteps , step , reform , failed ;
      real maxstep, endtime, ratio, tstep ;
      int  isub, nsub, new_nsub ;
      double exp ;

         time = time_data.time ;

         output_phase(0) ;

         if(time_data.ntime_grp == 0)  return ; 
         time_data.elastic = 0 ;

         time_data.do_slip = 0 ;
         time_data.next_slip = time + time_data.slip_interval ;
              
         printf("Starting time-stepping algorithm...\n") ;

         force.incr_rhs = (real * ) calloc(fe_sys.neq,sizeof(real)) ;
         if (force.incr_rhs == NULL) attempt = R_MEM ;
         completion() ;

         step_count = 0 ;

for ( time_group = 0 ; time_group < time_data.ntime_grp ; time_group++)
           {
             dt    = *(time_data.delt + time_group ) ;
             alpha = *(time_data.alpha + time_group ) ;
		     endtime = *(time_data.endtime + time_group ) ;
             time_data.nsteps[time_group] = 0;
		     time_data.nsub = 1 ;

		     step = -1 ;
		     while (time_data.time < endtime)
		       {
                time_data.nsteps[time_group]++;
			    ++step ;
			    step_count++ ;
			    tstep = dt/(real)time_data.nsub ;
                            time += tstep ;
                            time_data.dt = tstep ;
                            time_data.alpha_delt = tstep * alpha ;
                            time_data.time = time ;
                            reform = NO ;
                            time_data.do_slip = 0 ;
                            if(time >= time_data.next_slip)
                               {
                                time_data.do_slip = 1 ;
                          printf("Scheduled fault slip occuring at t = %f \n",time) ;
                                time_data.next_slip += time_data.slip_interval ;
                               }
                            if( time_data.do_slip || 
                               ((alpha != ZERO) && (step % time_data.nreform == 0)) )
                               {
                                clear_stiff() ;
                                elgrp_loop( FORMS_STEP ) ;
                                reform = YES ;

                /* Whenever we reform the stiffness, go ahead and
                   smooth the pressures as well.  But be careful --
                   this way, no smoothing will ever occur in the
                   explicit time stepping case (alpha = 0)          */
/*  - disabled -
                                if( fe_sys.psmooth > 0 )
                                   {
                                    clr_smooth( fe_sys.psmooth ) ;
                                    elgrp_loop( SMOOTH_BEGIN ) ;
                                    do_smooth() ;
                                    elgrp_loop( SMOOTH_END ) ;
                                    fe_sys.psmooth = 2 ;
                                   }
 - disabled - */
                               }
                            clear_real(global.del_displ,
                                        fe_sys.numnp*fe_sys.ndof) ;
                            node_load(time,reform) ;

			    time_data.max_strain = ZERO ;

                            elgrp_loop( RHS_STEP ) ;

			    if (time_data.max_strain > ZERO)
			    {
			    
                  ratio = time_data.max_strain * 5.0 * time_data.nsub ;
                  exp = log(ratio) / log(TWO) ;
                  exp = (double) ( (int) exp ) ;
			      new_nsub = MIN ((int) (.01+pow(2.0,exp+1.0)), MAXSUB) ;
			      new_nsub = MAX (1, new_nsub) ;

			    }

			    else
			    {
			      new_nsub = time_data.nsub ;
			    }

			    if (new_nsub != time_data.nsub)
			    {
                 if( !failed )
			       time -= tstep ;
			     time_data.time = time ;
                 printf("Changing nsub from %d to %d at time = %f\n", 
			     time_data.nsub, new_nsub, time) ;
			      time_data.nsub = new_nsub ;
                              step = (-1) ;
			      continue ;    /* roll back a time step */
			    }

                            
                 if(fe_sys.solver == PCG)
                     solver( ITER ) ;
                 else if(reform)
                     solver( FACBACK ) ;
                 else
                     solver( BACK ) ;
                               
                  accumulate() ;
                  elgrp_loop( V_STRESS ) ;

                  output_phase(0) ;

                  if((step_count+1) % time_data.nbackup == 0)
                       check_monitor() ;
                  fflush( stdout ) ;


		     }     /* end loop within time group */
             }   /* end loop over time groups  */
                   
             }
             
/*************************** end of time_step ********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   clear_stiff   ***************
*/

/*  INTERFACE:   */

     clear_stiff(
                )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine clear_stiff
** clear_stiff nulls the stiffness matrix memory.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

{
    switch(fe_sys.solver)
       {
        case DIRECT:
         clear_real(stiff.matrix,stiff.size) ;
         break;

         
        case PCG:
         clear_real(pcg.precond,pcg.neq) ;
         break;
       }

}
/*************************** end of clear_stiff ******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   elgrp_loop   ***************
*/

/*  INTERFACE:   */

     elgrp_loop(
                int  task      /* which element-level task to perform */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine elgrp_loop
** elgrp_loop invokes a particular task that relies on 
** element structures and affects all the elements in 
** every element group.
**
** Supported tasks are
** GENERATE: read or generate element data, fill ien, lm arrays
** OUTPUT: compute and write out element centroid stresses
** SHAPE: compute parent-space shape functions and gradients
** FORMS_ELAS: compute elastic stiffness matrix
** FORMS_STEP: compute VE single-step stiffness matrix
** RHS_ELAS: compute elastic right-hand side vector for FE problem
** RHS_STEP: compute VE right-hand side vector for FE problem
** BC_ELAS:  compute boundary-condition terms for elastic FE problem
** BC_STEP: compute boundary-condition terms for VE FE single step
** COL_HT: compute the column heights for profile stiffness storage
** E_STRESS: compute the stresses from elastic solution
** V_STRESS: compute the stresses from a VE step solution
** DUMP: dump information needed for restarts
** RESTORE: read and restore solution vector and info for a restart
** FAIL_CHECK: check for stress-driven fault failure
** SMOOTH_BEGIN: perform pressure smoothing set-up
** SMOOTH_END: apply pressure-smoothing function from coefficients
** REORDER: analyze adjacency, find profile-optimizing node permutation
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

  {
   int  i , result ;
   
   for(i=0 ; i<global.n_group ; i++)
      {
       switch( task )
               {
                case GENERATE:
                    gen_element( &groups[i] ) ;
                    break ;
                case OUTPUT:
                    el_output( &groups[i] ) ;
                    break ;
                case SHAPE:
                    p_shape( &groups[i] ) ;
                    completion() ;
                    break ;
                case FORMS_ELAS:
                    form_stiff( &groups[i] , ELASTIC ) ;
                    completion() ;
                    break ;
                case FORMS_STEP:
                    form_stiff( &groups[i] , VISCO ) ;
                    completion() ;
                    break ;
                case RHS_ELAS:
                    form_rhs( &groups[i] , ELASTIC ) ;
                    completion() ;
                    break ;
                case RHS_STEP:
                    form_rhs( &groups[i] , VISCO ) ;
                    completion() ;
                    break ;
                case BC_ELAS:
                    form_bc( &groups[i] , ELASTIC ) ;
                    completion() ;
                    break ;
                case BC_STEP:
                    form_bc( &groups[i] , VISCO ) ;
                    completion() ;
                    break ;
                case COL_HT:
                    colht( &groups[i] , &stiff ) ;
                    completion() ;
                    break ;
                case E_STRESS:
                    form_stress(&groups[i] , ELASTIC) ;
                    break ;
                case V_STRESS:
                    form_stress(&groups[i] , VISCO) ;
                    break ;
                case ESPROD:
                    estiffprod( &groups[i] ) ;
                    break ;
                case DUMP:
                    el_save(&groups[i] , DUMP) ;
                    break ;
                case RESTORE:
                    el_save(&groups[i] , RESTORE) ;
                    break ;
/*  - disabled -
                case SMOOTH_BEGIN:
                    form_smooth( &groups[i] , fe_sys.psmooth ) ;
                    completion() ;
                    break ;
                case SMOOTH_END:
                    apply_smooth( &groups[i] ) ;
                    completion() ;
                    break ;
 - disabled - */
                case REORDER:
                    reorder( &groups[i] );
                    completion() ;
                    break;
                default:
                    attempt = BAD_TASK ;
                    completion() ;
                    break ;
               }

      }  /* end loop */
  }
/*************************** end of elgrp_loop *******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   node_load   ***************
*/

/*  INTERFACE:   */

     node_load(
               real  time ,   /* current simulation time */
               int reform     /* flag indicating stiffness reform status */
              )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine node_load
** node_load applies node-based forces that may accumulate
** with time.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     real  test, val ;
     int  i, k, n , id , ndof , numnp , new_step ;
     int pgnode; /* one-based permuted node numbers */
    
    ndof = fe_sys.ndof ;
    numnp = fe_sys.numnp ;
    if(time == ZERO)          /* the ELASTIC case... */
       {
        /*  put in free nodal forces, if any  */
        for(i=0 ; i<numnp ; i++)
           {
            for(k=0 ; k<ndof ; k++)
               {
                pgnode = global.permut[i+1];
                id = *(global.id_pointer+(pgnode-1)*ndof+k) ;
                val = *(global.forv+i*ndof+k) ;
                if(val != ZERO && id != (-1))
                  *(force.full_rhs + id) = val ;
               }
           }
        return ;
       }
    
             /* the TIME-STEPPING case... */
    
     new_step = 0 ;
     if(fe_sys.nrates == 0)
       {
        clear_real(force.full_rhs,fe_sys.neq) ;
        clear_real(global.forv , fe_sys.numnp*fe_sys.ndof) ;
        return ;
       }
     for(n=0 ; n<fe_sys.nrates ; n++)
        {
         test = *(global.bctime + n) ;
         if( test >= ZERO && test <= time )
            {
             new_step = 1 ;
             move_real(global.rate + n*fe_sys.numnp*fe_sys.ndof
                        , global.forv , fe_sys.numnp*fe_sys.ndof) ;
             *(global.bctime + n) = (-ONE) ;
             break ;
            }
        }

     if(reform || new_step)  /* only redo the rhs if the bc or stiffness have changed */
       {
        clear_real(force.incr_rhs,fe_sys.neq) ;
        elgrp_loop( BC_STEP ) ;
       }

        for(i=0 ; i<numnp ; i++)
           {
            for(k=0 ; k<ndof ; k++)
               {
                pgnode = global.permut[i+1];
                id = *(global.id_pointer+(pgnode-1)*ndof+k) ;
                val = *(global.forv+i*ndof+k) * time_data.dt ;
                if(val != ZERO && id != (-1))
                  *(force.incr_rhs + id) = val ;
               }
           }

     move_real(force.incr_rhs , force.full_rhs , fe_sys.neq) ;

    }
/*************************** end of node_load ********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   accumulate   ***************
*/

/*  INTERFACE:   */

     accumulate(
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine accumulate
** accumulate updates the total displacement based on the 
** single-step increment.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     COUNTER   i ;

     for(i=0 ; i<fe_sys.numnp*fe_sys.ndof ; i++)
        {
         *(global.displ + i) += *(global.del_displ + i) ;
        }     
    }

/*************************** end of accumulate *******************************/ 


/*
** Routine clr_smooth
** clr_smooth nulls the smoothing coefficients for 
** pressure smoothing.
**/
/*  disabled ...
 clr_smooth( ypflag )
   int  ypflag ;
    {
     COUNTER   i ;

     for(i=0 ; i<fe_sys.numnp ; i++)
        {
         *(global.pnode + i) = ZERO ;
        }     
     if( ypflag == 1 )
         {
          for(i=0 ; i<fe_sys.numnp ; i++)
             {
              *(global.yp + i) = ZERO ;
             }     
         }
    }
 ... disabled*/
/********************** end of function ****************************/ 

/*
** Routine do_smooth
** do_smooth applies the smoothing coefficients for 
** pressure smoothing
**/
/*    -disabled -
 do_smooth()
    {
     COUNTER   i ;

     for(i=0 ; i<fe_sys.numnp ; i++)
        {
         *(global.pnode + i) /= *(global.yp + i) ;
        }     
    }
*/
/********************** end of function ****************************/ 

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   completion   ***************
*/

/*  INTERFACE:   */

     completion(
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine completion
** completion syncronizes multiple processors. 
** Parallel vestige; keep for now.
** Keeps track of incurred errors in each processor.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     COUNTER   i ;
     static int   all ;
     
     all = attempt ;
     if(all)
        {
         printf("An error occurred:\n") ;
         printf(" error code = %d\n",attempt) ;
         exit(1) ;
        }
    }

/*************************** end of completion *******************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   wrt_save   ***************
*/

/*  INTERFACE:   */

     wrt_save(
             )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine wrt_save
** wrt_save dumps state to disk for restarts.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     COUNTER  i , k , index , ndof , numnp ;
     static real  outbuf[4] ;
     real  time ;
         
     ndof = fe_sys.ndof ;
     numnp = fe_sys.numnp ;
     time = time_data.time ;
     
     if((dump_file = fopen(time_data.savefile,"w")) == NULL)
          {
           printf("Dump file couldn't be opened!\n") ;
           return ;
          }

     printf("Saving simulation at time = %f to disk...\n",time) ;
     fprintf(dump_file,"%f\n",time) ;
     
     for(i=0 ; i<numnp ; i++)
                 {
                  index = i ;
        
                      for(k=0 ; k<ndof ; k++)
                         {
                          outbuf[k] =
                             *(global.displ + ndof*index + k) ;
                         }
                            
                  fprintf(dump_file,"%d",i+1) ;
                  for(k=0 ; k<ndof ; k++)
                     {
                      fprintf(dump_file,"  %f",outbuf[k]) ;
                     }
                  fprintf(dump_file,"\n") ;
        
                 }        
             fflush(dump_file) ;
             elgrp_loop(DUMP) ;
             fclose(dump_file) ;
             printf("Simulation dump complete.\n") ;
    }

/*************************** end of wrt_save *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   vrestart   ***************
*/

/*  INTERFACE:   */

     vrestart(
             )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** vrestart restores state from previous dump and 
** continues simulation.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     COUNTER  i , k , n , index , ndof , numnp ;
     static real  outbuf[4] ;
     real  time ;
         
     ndof = fe_sys.ndof ;
     numnp = fe_sys.numnp ;
     time = time_data.time ;
     
     if((dump_file = fopen(time_data.startfile,"r")) == NULL)
          {
           printf("Restart file %s couldn't be opened!\n",time_data.startfile) ;
           exit(1) ;
          }

     fscanf(dump_file,"%lf\n",&time) ;
     time_data.time = time ;
     printf("Restarting simulation at time = %f ...\n",time) ;
     
     for(i=0 ; i<numnp ; i++)
                 {
                  fscanf(dump_file,"%d",&n) ;
                  for(k=0 ; k<ndof ; k++)
                     {
                      fscanf(dump_file,"%lf",outbuf+k) ;
                     }
                  index = i ;
                      for(k=0 ; k<ndof ; k++)
                         {
                          *(global.displ + ndof*index + k) =
                             outbuf[k] ;
                         }
                 }        

             elgrp_loop(RESTORE) ;
             fclose(dump_file) ;
             printf("Simulation restoration complete.\n") ;
    }

/*************************** end of vrestart *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   el_save   ***************
*/

/*  INTERFACE:   */

     el_save(
             GROUP   *grp_ptr ,       /* pointer to current element group */
             int     code             /* code specifying which action */
            )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine el_save
** el_save dumps or restores element data.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      real   str_temp ;
      int     numel , nint , nel , n , str_size ,
              ipt , j ;

      ELEMENT_DATA   *el_pt ;

if( code == DUMP )
     {
      fflush(dump_file) ;
      numel = (grp_ptr->el_info)->numel ;
      str_size = (grp_ptr->el_info)->stress_size ;
      
      switch((grp_ptr->el_info)->type)
          {
           case BILIN:
           case SERENDIP:
           case TET:

               for ( nel = 0 ; nel < numel ; nel++)
                   {
                    fflush(dump_file) ;

                    n = nel + 1 ;
                    
                    el_pt = grp_ptr->el_data + n - 1 ;
                    nint = el_pt->nint ;

                    for (ipt = 0 ; ipt < nint ; ipt++)
                       {
                        for(j=0 ; j<str_size ; j++)
                           fprintf( dump_file,"%f  ",
                            *((el_pt -> stress) + ipt*str_size + j) ) ;
                        fprintf(dump_file,"\n") ;
                       }
                   }
               fflush(dump_file) ;
               break ;
               
          }
     }

else if( code == RESTORE )
     {
      numel = (grp_ptr->el_info)->numel ;
      str_size = (grp_ptr->el_info)->stress_size ;
      
      switch((grp_ptr->el_info)->type)
          {
           case BILIN:
           case SERENDIP:

               for ( nel = 0 ; nel < numel ; nel++)
                   {
                    n = nel + 1 ;
                    el_pt = grp_ptr->el_data + n - 1 ;
                    nint = el_pt->nint ;
                    for (ipt = 0 ; ipt < nint ; ipt++)
                       {
                        for(j=0 ; j<str_size ; j++)
                           {
                            fscanf( dump_file,"%lf",
                                    &str_temp ) ;
                            if(n != 0)
                               {
                                el_pt = grp_ptr->el_data + n - 1 ;
                                *((el_pt -> stress) + ipt*str_size + j) =
                                  str_temp ;
                               }
                           }
                       }
                   }
               break ;
               
          }
     }

    }

/**************************** end of el_save *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   check_monitor   ***************
*/

/*  INTERFACE:   */

     check_monitor(
                  )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine check_monitor
** check_monitor allows user run intervention via strings stored
** in the file monitor.fem.
**
** Supported options:
** OK: close the dump file (?)
** KILL: close files and abort
** SAVE_STOP: write state, then abort
** SAVE_GO:  write dumps as requested in input, and continue
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

    {
     int  i , n ;
     
     printf("Simulation running normally at time = %f ; step size = %f\n",
             time_data.time , time_data.dt) ;
     if((dump_file = fopen("monitor.fem","r")) == NULL)
          {
           return ;
          }

     fscanf(dump_file,"%s",line) ;
     n = NCODES ;
     for(i=0 ; i<NCODES ; i++)
        {
         if(strcmp(line , mon_codes[i]) == 0 )
            {
             n = i ;
             break ;
            }
        }

     switch( n )
        {
         case OK:
            fclose(dump_file) ;
            return ;
            break ;

         case KILL:
            fclose(in_file) ;
            fclose(out_file) ;
            fclose(dump_file) ;
            printf("Run aborted by monitor request.\n") ;
            exit(0) ;
            break ;

         case SAVE_STOP:
            fclose(dump_file) ;
            if(time_data.sav_state)
               wrt_save() ;
            else
               {
                sprintf(time_data.savefile,"save.def") ;
                wrt_save() ;
               }
            fclose(in_file) ;
            fclose(out_file) ;
            printf("Run saved and halted by monitor request.\n") ;
            exit(0) ;
            break ;

         case SAVE_GO:
            fclose(dump_file) ;
            if(time_data.sav_state)
               wrt_save() ;
            else
               {
                sprintf(time_data.savefile,"save.def") ;
                wrt_save() ;
               }
            printf("Run backed up and continuing...\n") ;
            return ;
            break ;

         default:
            fclose(dump_file) ;
            return ;
            break ;
        }
    }

/************************** end of check_monitor *****************************/ 
