/*
***                           File stiff.c                        ***
***                        GeoFEST version 4.2                    ***
*** Copyright (c) 2002, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This file contains modules which construct the finite
*** element stiffness matrix and the right-hand-side "force" vector.
***
***          == form_stiff ==
***          == form_rhs ==
***          == form_bc ==
***          == form_slip ==
***          == adjust_bc ==
***          == lame_form ==
***          == force_form ==
***          == surf_form ==
***          == shape ==
***          == dotsh ==
***          == adfldp ==
***          == p_shape ==
***          == form_smooth == - disabled -
***          == addsmooth == - disabled -
***          == apply_smooth == - disabled -
***/

/*
***GeoFEST version 4.2***
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining 
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
$Log: stiff.c,v $
Revision 1.1.1.1  2002/07/12 19:05:14  jwp
Commiting GeoFEST to CVS at version 4.2

*/

#include <stdio.h>
#include "mydefs.h"
#include "errcodes.h"
#define EXTERN extern
#include "finel.h"
#include <math.h>

#define SH(i,n) *(sh + nsdp1*(n) + (i))
#define ES(i,j) *(es + nee*(j) + (i))

#define PSH_BILIN(i,enode,point)\
 *(psh_b + 4*nsdp1*(enode) + nsdp1*(point) + (i))

#define PSH_SEREN(i,enode,point,rule)\
 *(psh_s + 18*nsdp1*(enode) + 9*nsdp1*(rule) + nsdp1*(point) + (i))

#define PSH_TET(i,enode,point)\
 *(psh_t + 4*nsdp1*(enode) + nsdp1*(point) + (i))

#define PSH_TRILIN6(i,enode,point)\
 *(psh_h6 + 6*nsdp1*(enode) + nsdp1*(point) + (i))

#define PSH_TRILIN8(i,enode,point)\
 *(psh_h8 + 8*nsdp1*(enode) + nsdp1*(point) + (i))

   /* Gaussian quadrature points (without scale factors) and weights */

#define WT1  0.78867513
#define WT2  0.21132487
#define RTET1 0.138196601125011
#define RTET2 0.585410196624969

 /* 1/24 */
#define WTET 0.041666666666667               
 /* 4/3 */
#define WHEX 1.333333333333333               

static real
   x_2x2[4] = {-ONE  , ONE  , ONE  , -ONE } ,
   y_2x2[4] = {-ONE  ,-ONE  , ONE  ,  ONE } ,
   wt_2x2[4] = { ONE  , ONE  , ONE  ,  ONE} ,

   x_2x2x2[8] = {-ONE  , ONE  , ONE  , -ONE , -ONE  , ONE  , ONE , -ONE } ,
   y_2x2x2[8] = {-ONE  , -ONE  , ONE  , ONE , -ONE  , -ONE  , ONE , ONE } ,
   z_2x2x2[8] = {-ONE  , -ONE  , -ONE  , -ONE , ONE  , ONE  , ONE , ONE } ,
   wt_2x2x2[8] = {ONE  , ONE  , ONE  , ONE , ONE  , ONE  , ONE , ONE } ,

   x_3x3[9] =
    {-ONE  , ONE  , ONE  , -ONE , ZERO  ,  ONE , ZERO  , -ONE , ZERO} ,
   y_3x3[9] =
    {-ONE  ,-ONE  , ONE  ,  ONE , -ONE  , ZERO ,  ONE  , ZERO , ZERO} ,
   wt_3x3[9] =
    { 25.0/81.0 , 25.0/81.0 , 25.0/81.0 , 25.0/81.0 , 40.0/81.0 ,
      40.0/81.0 , 40.0/81.0 , 40.0/81.0 , 64.0/81.0               } ,

   x_tet[4] = {RTET1, RTET2, RTET1, RTET1} ,
   y_tet[4] = {RTET1, RTET1, RTET2, RTET1} ,
   z_tet[4] = {RTET1, RTET1, RTET1, RTET2} ,
   wt_tet[4] = {WTET,WTET,WTET,WTET} ;

static real
   x_tri[6] = {-ONE, ZERO, ZERO, ONE, ZERO, ZERO} ,
   y_tri[6] = {ZERO, -ONE, ZERO, ZERO, ONE, ZERO} ,
   z_tri[6] = {ZERO, ZERO, -ONE, ZERO, ZERO, ONE} ,
   wt_tri[6] = {WHEX,WHEX,WHEX,WHEX,WHEX,WHEX} ,
   ra[8] = {-0.5, 0.5, 0.5,-0.5, -0.5, 0.5,0.5,-0.5} ,
   sa[8] = {-0.5,-0.5, 0.5, 0.5, -0.5,-0.5,0.5, 0.5} ,
   ta[8] = {-0.5,-0.5,-0.5,-0.5,  0.5, 0.5,0.5, 0.5} ;

static real
  psh_b[48] ,
   /* parent space shape functions for 4-node bilinear quad */
  psh_s[432]   ,
   /* parent space shape functions for 8-node serendipity quad */
  psh_t[64]    ,
   /* parent space shape functions for 4-node tet */
  psh_h6[192]    ,
   /* parent space shape functions for 8-node hexahedron (6-pt rule) */
  psh_h8[256]    ;
   /* parent space shape functions for degen 8-node hexahedron (8-pt rule) */

static real
   el_stiff[576] , el_rhs[24] , temp[24] , sh_temp[256] , det_temp[9] ;

static int  el_diag[24] =
            { 0 , 2 , 5 , 9 , 14 , 20 , 27 , 35 , 44 , 54 , 65 , 77 ,
             90 , 104 , 119 , 135 , 152 , 170 , 189 , 209 , 230 ,
             252 , 275 , 299 } ;

real  dotsh() ;
real adfldp() ;

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_stiff   ***************
*/

/*  INTERFACE:   */

     form_stiff(
                GROUP   *grp_ptr ,   /* pointer to current element group */
                int     code         /* code specifying elastic or visco */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_stiff
** form_stiff computes the element-wise stiffness array for the elt group.
** "code" indicates if this is for an elastic problem or a VE step.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      char CVS_Id[] = "@(#) CVS id $Id: stiff.c,v 1.1.1.1 2002/07/12 19:05:14 jwp Exp $, Revision: " REVNO;
      int     numel , nel , nsd , nsdp1 ;

      ELEMENT_DATA   *el_pt ;


      nsd = fe_sys.nsd ;
      nsdp1 = nsd + 1 ;
      numel = (grp_ptr->el_info)->numel ;
      
      switch((grp_ptr->el_info)->type)
          {
           case TET:
           case BILIN:
           case SERENDIP:
           case TRILIN:
              for ( nel = 0 ; nel < numel ; nel++)
                   {
                    el_pt = grp_ptr->el_data + nel ;
                    lame_form( code , el_stiff , grp_ptr->el_info ,
                              grp_ptr->el_mat, el_pt ) ;

                    addstiff( &stiff , el_stiff ,
                              grp_ptr->el_info , el_pt ) ;
                   }
               break ;
               
          }
    }

/************************** end of form_stiff ********************************/ 

 
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_rhs   ***************
*/

/*  INTERFACE:   */

     form_rhs(
              GROUP   *grp_ptr ,   /* pointer to current element group */
              int     code         /* code specifying elastic or visco */
             )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_rhs
** form_rhs computes the right-hand-side vector for the FE problem.
** "code" indicates if this is for the elastic problem or a VE step.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      int    n , numel , nel , ndof , numsuf , nfnode , side ,
             nsd , nsdp1 , node ;
      int   *surf_list , *slip_list ;
      real  *surf_val , *slip_val , *trac , *slip ;
      ELEMENT_DATA   *el_pt ;


      nsd = fe_sys.nsd ;
      nsdp1 = nsd + 1 ;
      numel = (grp_ptr->el_info)->numel ;
      numsuf = (grp_ptr->el_info)->numsuf ;
      nfnode = (grp_ptr->el_info)->nfnode ;
      ndof = fe_sys.ndof ;

      switch((grp_ptr->el_info)->type)
          {
           case TET:
           case BILIN:
           case SERENDIP:
           case TRILIN:
               for ( nel = 0 ; nel < numel ; nel++)
                   {
                    el_pt = grp_ptr->el_data + nel ;
                    force_form( code , el_rhs , grp_ptr->el_info ,
                              grp_ptr->el_mat, el_pt ) ;
                    addfor( force.full_rhs , el_rhs , grp_ptr->el_info ,
                            el_pt ) ;
                   }

               if(code == ELASTIC)

                if( (code == ELASTIC) && time_data.elastic)
                   {
                    surf_list = (grp_ptr->el_info) -> surf_list ;
                    surf_val = (grp_ptr->el_info) -> surf_trac ;

/************************************************************************/
/*
surf_list[] is a numsuf-by-2-deep array of ints containing:
        [0] = global element number
        [1] = face number; 
    example: face 3 is the side comprised of nodes 1,2 and 4

surf_val[] is a numsuf-by-ndof-deep array of reals containing:
         [0] = x-component of surface traction
         [1] = y-component of surface traction
         [2] = z-component of surface traction
*/
/************************************************************************/

                    for(n=0 ; n<numsuf ; n++)
                       {
                        nel = *(surf_list + 2*n) ;
                        side = (int) *(surf_list + 2*n + 1) ;
                        trac = (surf_val + ndof*n) ;
                        el_pt = grp_ptr->el_data + nel ;
                        surf_form( code , el_rhs , grp_ptr->el_info ,
                                   side , trac , el_pt ) ;
                        addfor( force.full_rhs , el_rhs ,
                                grp_ptr->el_info , el_pt ) ;
                       }
                   }

                if( time_data.do_slip )
                   {
                    slip_list = (grp_ptr->el_info) -> slip_list ;
                    slip_val = (grp_ptr->el_info) -> slip_val ;
                    for(n=0 ; n<nfnode ; n++)
                       {
                        nel = (int) *(slip_list + 2*n) ;
                        node = (int) *(slip_list + 2*n + 1) ;
                        slip = (slip_val + ndof*n) ;
                        el_pt = grp_ptr->el_data + nel ;
                        form_slip( code , el_rhs , el_stiff , grp_ptr->el_info ,
                                   nel , node , slip , el_pt , grp_ptr->el_mat ) ;
                        addfor( force.full_rhs , el_rhs ,
                                grp_ptr->el_info , el_pt ) ;
                       }
                   }

               break ;
               
          }
    }

/************************** end of form_rhs **********************************/ 
 

/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_bc   ***************
*/

/*  INTERFACE:   */

     form_bc(
             GROUP   *grp_ptr ,   /* pointer to current element group */
             int     code         /* code specifying elastic or visco */
            )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_bc
** form_bc computes the boundary condition terms for stiffness and rhs
** due to imposed conditions
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      int     numel , nen , nee , nel , i , j ,
              active , node , ndof , nsd , nsdp1 ;
      ELEMENT_DATA   *el_pt ;


      nsd = fe_sys.nsd ;
      nsdp1 = nsd + 1 ;
      numel = (grp_ptr->el_info)->numel ;
      nen = (grp_ptr->el_info)->nen ;
      nee = (grp_ptr->el_info)->nee ;
      ndof = fe_sys.ndof ;

      switch((grp_ptr->el_info)->type)
          {
           case TET:
           case BILIN:
           case SERENDIP:
           case TRILIN:
               for( nel = 0 ; nel < numel ; nel++)
                   {
                    el_pt = grp_ptr->el_data + nel ;
                    if( !(el_pt->is_bc) )  continue ;
                    
                    lame_form( code , el_stiff , grp_ptr->el_info ,
                               grp_ptr->el_mat, el_pt ) ;

                    for ( i = 0 ; i < nee ; i++ )
                         el_rhs[i] = ZERO ;
                        
                        active = 0 ;
                        for(j=0 ; j<nen ; j++)
                           {
                            node = *(el_pt->ien + j) - 1 ;
                            for(i=0 ; i<ndof ; i++)
                               {
                                if( *(el_pt->lm + ndof*j + i) != (-1) ||
                                    *(global.forv + node*ndof + i) ==
                                    ZERO )
                                     {
                                      *(temp + ndof*j + i) = ZERO ;
                                     }
                                else
                                     {
                                      *(temp + ndof*j + i) =
                                        *(global.forv + node*ndof + i) ;
                                      active = 1 ;
                                     }
                               }
                           }
                        if(code == VISCO)
                           {
                            for(i=0 ; i<nee ; i++)
                               *(temp + i) *= time_data.dt ;
                           }
                        if( active )
                           adjust_bc( temp , el_rhs , el_stiff , nee ) ;

                    if(code == VISCO)
                       addfor( force.incr_rhs , el_rhs , grp_ptr->el_info ,
                               el_pt ) ;
                    else
                       addfor( force.full_rhs , el_rhs , grp_ptr->el_info ,
                               el_pt ) ;
                   }

               break ;
               
          }
    }

/************************** end of form_bc ***********************************/ 

 
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   form_slip   ***************
*/

/*  INTERFACE:   */

     form_slip(
               int  code ,     /* code specifying elastic or visco */
               real  *rhs ,   /* temp array to hold element rhs contrib */
               real  *stiff , /* temp array to hold el stiffness contrib */
               ELEMENT_INFO  *info ,   /* ptr to element info struct */
               int  nel ,       /* current element number */
               int  node ,      /* current node number */
               real  *slip ,   /* array containing fault slip amplitudes */
               ELEMENT_DATA  *el_pt ,  /* ptr to current element struct */
               ELEMENT_MAT   *mat_pt  /* ptr to el material prop struct */
              )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine form_slip
** form_slip computes the split-nodes fault offsets and their influence 
** on the finite element stiffness and RHS.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      int  i , nee , ndof , nen , j , active ;
     
      nen     = info->nen ;
      nee     = info->nee ;
      ndof     = fe_sys.ndof ;
      
      for (  i = 0 ; i < nee ; i++ )
          *(rhs + i) = ZERO ;
      active = 0 ;
          
      lame_form( code , stiff , info , mat_pt , el_pt ) ;

      for(j=0 ; j<nen ; j++)
         {
          if( *(el_pt->ien + j) == node )
             {
              for(i=0 ; i<ndof ; i++)
                  {
                   *(temp + ndof*j + i) = *(slip + i) ;
                  }
              active = 1 ;
             }
          else
             {
              for(i=0 ; i<ndof ; i++)
                  {
                   *(temp + ndof*j + i) = ZERO ;
                  }
             }
          }

      if( active )
         adjust_bc( temp , rhs , stiff , nee ) ;

         }

/************************** end of form_slip *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   adjust_bc   ***************
*/

/*  INTERFACE:   */

     adjust_bc(
               real  *disp ,   /* array containing fault slip amplitudes */
               real  *rhs ,   /* temp array to hold element rhs contrib */
               real  *es ,    /* temp array to hold el stiffness contrib */
               int   nee        /* number of element equations */
              )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine adjust_bc
** adjust_bc computes the necessary terms that modify the right-hand-side
** due to imposed displacements.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    int i , j ;
    
    for(j=0 ; j<nee ; j++)
       {
        for(i=0 ; i<=j ; i++)
           {
            if( disp[i] == ZERO ) continue ;
            
            rhs[j] -= ES(i,j) * disp[i] ;
           }
            
        if( j < (nee-1) )
           {
            for(i=(j+1) ; i<nee ; i++)
               {
                if( disp[i] == ZERO ) continue ;
            
                rhs[j] -= ES(j,i) * disp[i] ;
               }
           }
       }
   }
/************************** end of adjust_bc *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   lame_form   ***************
*/

/*  INTERFACE:   */

     lame_form(
               int  code ,     /* code specifying elastic or visco */
               real  *es ,    /* temp array to hold el stiffness contrib */
               ELEMENT_INFO  *info ,   /* ptr to element info struct */
               ELEMENT_MAT   *mat_pt ,  /* ptr to el material prop struct */
               ELEMENT_DATA  *el_pt    /* ptr to current element struct */
              )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine lame_form
** lame_form computes the volumetric stiffness term for one element,
** that is based on the constitutive volume constants. 
** Results are stored in the element stiffness array.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      real  w_det , lam_2mu , lam_el  , mu_el ;
      real  db11 , db21 , db31 , db41 , db51 , db61 ,
             db12 , db22 , db32 , db42 , db52 , db62 ,
             db13 , db23 , db33 , db43 , db53 , db63 ;
      real  bcol[6] , *shape_ptr , *det_ptr , *sh , *dbar , *wt ;
      int    i , j , ipt , it2 , jt2 , it3 , jt3 ,
             nint , nen , nee , nee_sq , ndof , str_size , dbar_size ,
             save_shape , type , nsd , nsdp1 ;
             

      nsd = fe_sys.nsd ;
      nsdp1 = nsd + 1 ;
      nint = el_pt->nint ;
      lam_el  = *(mat_pt->lambda + el_pt->mat) ;
      mu_el   = *(mat_pt->mu + el_pt->mat) ;
      lam_2mu = lam_el + TWO * mu_el ;
      nen     = info->nen ;
      nee     = info->nee ;
      nee_sq = info->el_size ;
      type = info->type ;
      ndof     = fe_sys.ndof ;
      str_size = info->stress_size ;
      dbar_size = info->dbar_size ;
      save_shape = fe_sys.save_shape;
      switch( nint )
         {
          case 4:
             wt = (type == BILIN) ? wt_2x2 : wt_tet ;
             break ;
          case 6:
             wt = wt_tri ;
             break ;
          case 8:
             wt = wt_2x2x2 ;
             break ;
          case 9:
             wt = wt_3x3 ;
             break ;
         }
      if( !(*(mat_pt->plastic + el_pt->mat)) )
         code = ELASTIC ;

      if( !save_shape )
         {
          shape_ptr = sh_temp ;
          det_ptr = det_temp ;
          shape( el_pt , info , shape_ptr , det_ptr , type , nen , nint , 1 ) ;
         }
      else
         {
          shape_ptr = el_pt -> shape ;
          det_ptr = el_pt -> deter ;
         }

      for (  i = 0 ; i < nee_sq ; i++ )
          *(es + i) = ZERO ;

      if( code == VISCO )
      {

         form_dbar(el_pt  ,  mat_pt  , info) ;

      }
          
      for (ipt = 0 ; ipt < nint ; ipt++)
           {
            sh = shape_ptr + ipt*nen*nsdp1 ;
            w_det = wt[ipt] * *(det_ptr + ipt) ;
            dbar = (el_pt -> dbar) + ipt*dbar_size ;
            
            switch( ndof )
               {
                case 1:
                       if( code == ELASTIC )
                          {
                           for(j=0;j<nen;j++)
                              {
                               for(i=0;i<=j;i++)
                                  {
                                   ES(i,j) += w_det*mu_el*
                                     (SH(0,i)*SH(0,j)+SH(1,i)*SH(1,j)) ;
                                  }
                              }
                          }
                       else if( code == VISCO )
                          {
                           for(j=0;j<nen;j++)
                              {
                               bcol[0]  = w_det * SH(0,j) ;
                               bcol[1]  = w_det * SH(1,j) ;
                               full_back(dbar,bcol,el_diag,str_size) ;
                               db11 = bcol[0] ;
                               db21 = bcol[1] ;
                               for(i=0;i<=j;i++)
                                  {
                                   ES(i,j) +=
                                      SH(0,i)*db11+SH(1,i)*db21 ;
                                  }
                              }
                          }
                   break ;
                   
                case 2:
                       if( code == ELASTIC )
                          {
                           for(j=0;j<nen;j++)
                              {
                               jt2 = j+j ;
                               db11 = w_det * lam_2mu * SH(0,j) ;
                               db12 = w_det * lam_el * SH(1,j) ;
                               db21 = w_det * lam_el * SH(0,j) ;
                               db22 = w_det * lam_2mu * SH(1,j) ;
                               db31 = w_det * lam_el * SH(0,j) ;
                               db32 = w_det * lam_el * SH(1,j) ;
                               db41 = w_det * mu_el * SH(1,j) ;
                               db42 = w_det * mu_el * SH(0,j) ;
                               for(i=0;i<=j;i++)
                                  {
                                   it2 = i+i ;
                                   ES(it2 , jt2)     +=
                                      SH(0,i)*db11 + SH(1,i)*db41 ;
                                   ES(it2 , jt2+1)   +=
                                      SH(1,i)*db42 + SH(0,i)*db12 ;
                                   ES(it2+1 , jt2)   +=
                                      SH(0,i)*db41 + SH(1,i)*db21 ;
                                   ES(it2+1 , jt2+1) +=
                                      SH(1,i)*db22 + SH(0,i)*db42 ;
                                  }
                              }
                          }

                       else if( code == VISCO )
                          {
                           for(j=0;j<nen;j++)
                              {
                               jt2 = j+j ;
                               bcol[0]  = w_det * SH(0,j) ;
                               bcol[1]  = ZERO ;
                               bcol[2]  = ZERO ;
                               bcol[3]  = w_det * SH(1,j) ;
                               full_back(dbar,bcol,el_diag,str_size) ;
                               db11 = bcol[0] ;
                               db21 = bcol[1] ;
                               db31 = bcol[2] ;
                               db41 = bcol[3] ;
                               bcol[0]  = ZERO ;
                               bcol[1]  = w_det * SH(1,j) ;
                               bcol[2]  = ZERO ;
                               bcol[3]  = w_det * SH(0,j) ;
                               full_back(dbar,bcol,el_diag,str_size) ;
                               db12 = bcol[0] ;
                               db22 = bcol[1] ;
                               db32 = bcol[2] ;
                               db42 = bcol[3] ;
                               for(i=0;i<=j;i++)
                                  {
                                   it2 = i+i ;
                                   ES(it2 , jt2)     +=
                                      SH(0,i)*db11 + SH(1,i)*db41 ;
                                   ES(it2 , jt2+1)   +=
                                      SH(1,i)*db42 + SH(0,i)*db12 ;
                                   ES(it2+1 , jt2)   +=
                                      SH(0,i)*db41 + SH(1,i)*db21 ;
                                   ES(it2+1 , jt2+1) +=
                                      SH(1,i)*db22 + SH(0,i)*db42 ;
                                  }
                              }
                          }
                   break ;
                   
                case 3:
                       if( code == ELASTIC  && nsd == 2)
                          {
                           for(j=0;j<nen;j++)
                              {
                               jt3 = 3*j ;
                               db11 = w_det * lam_2mu * SH(0,j) ;
                               db12 = w_det * lam_el * SH(1,j) ;
                               db21 = w_det * lam_el * SH(0,j) ;
                               db22 = w_det * lam_2mu * SH(1,j) ;
                               db41 = w_det * mu_el * SH(1,j) ;
                               db42 = w_det * mu_el * SH(0,j) ;
                               for(i=0;i<=j;i++)
                                  {
                                   it3 = 3*i ;
                                   ES(it3 , jt3)     +=
                                      SH(0,i)*db11 + SH(1,i)*db41 ;
                                   ES(it3+1 , jt3)   +=
                                      SH(1,i)*db21 + SH(0,i)*db41 ;
                                   ES(it3+2 , jt3)    = ZERO ;
                                   ES(it3 , jt3+1)   +=
                                      SH(0,i)*db12 + SH(1,i)*db42 ;
                                   ES(it3+1 , jt3+1) +=
                                      SH(1,i)*db22 + SH(0,i)*db42 ;
                                   ES(it3+2 , jt3+1)  = ZERO ;
                                   ES(it3 , jt3+2)    = ZERO ;
                                   ES(it3+1 , jt3+2)  = ZERO ;
                                   ES(it3+2 , jt3+2) +=
                                      SH(0,i)*db42 + SH(1,i)*db41 ;
                                  }
                              }
                          }

                       else if( code == VISCO && nsd == 2)
                          {
                           for(j=0;j<nen;j++)
                              {
                               jt3 = 3*j ;
                               bcol[0]  = w_det * SH(0,j) ;
                               bcol[1]  = ZERO ;
                               bcol[2]  = ZERO ;
                               bcol[3]  = w_det * SH(1,j) ;
                               bcol[4]  = ZERO ;
                               bcol[5]  = ZERO ;
                               full_back(dbar,bcol,el_diag,str_size) ;
                               db11 = bcol[0] ;
                               db21 = bcol[1] ;
                               db31 = bcol[2] ;
                               db41 = bcol[3] ;
                               db51 = bcol[4] ;
                               db61 = bcol[5] ;
                               bcol[0]  = ZERO ;
                               bcol[1]  = w_det * SH(1,j) ;
                               bcol[2]  = ZERO ;
                               bcol[3]  = w_det * SH(0,j) ;
                               bcol[4]  = ZERO ;
                               bcol[5]  = ZERO ;
                               full_back(dbar,bcol,el_diag,str_size) ;
                               db12 = bcol[0] ;
                               db22 = bcol[1] ;
                               db32 = bcol[2] ;
                               db42 = bcol[3] ;
                               db52 = bcol[4] ;
                               db62 = bcol[5] ;
                               bcol[0]  = ZERO ;
                               bcol[1]  = ZERO ;
                               bcol[2]  = ZERO ;
                               bcol[3]  = ZERO ;
                               bcol[4]  = w_det * SH(0,j) ;
                               bcol[5]  = w_det * SH(1,j) ;
                               full_back(dbar,bcol,el_diag,str_size) ;
                               db13 = bcol[0] ;
                               db23 = bcol[1] ;
                               db33 = bcol[2] ;
                               db43 = bcol[3] ;
                               db53 = bcol[4] ;
                               db63 = bcol[5] ;
                               for(i=0;i<=j;i++)
                                  {
                                   it3 = 3*i ;
                                   ES(it3 , jt3)     +=
                                      SH(0,i)*db11 + SH(1,i)*db41 ;
                                   ES(it3+1 , jt3)   +=
                                      SH(1,i)*db21 + SH(0,i)*db41 ;
                                   ES(it3+2 , jt3)    =
                                      SH(0,i)*db51 + SH(1,i)*db61 ;
                                   ES(it3 , jt3+1)   +=
                                      SH(0,i)*db12 + SH(1,i)*db42 ;
                                   ES(it3+1 , jt3+1) +=
                                      SH(1,i)*db22 + SH(0,i)*db42 ;
                                   ES(it3+2 , jt3+1)  =
                                      SH(0,i)*db52 + SH(1,i)*db62 ;
                                   ES(it3 , jt3+2)    =
                                      SH(0,i)*db13 + SH(1,i)*db43 ;
                                   ES(it3+1 , jt3+2)  =
                                      SH(1,i)*db23 + SH(0,i)*db43 ;
                                   ES(it3+2 , jt3+2) +=
                                      SH(0,i)*db53 + SH(1,i)*db63 ;
                                  }
                              }
                          }
                    if( code == ELASTIC  && nsd == 3)
                          {
                           for(j=0;j<nen;j++)
                              {
                               jt3 = 3*j ;
                               db11 = w_det * lam_2mu * SH(0,j) ;
                               db12 = w_det * lam_el * SH(1,j) ;
                               db13 = w_det * lam_el * SH(2,j) ;
                               db21 = w_det * lam_el * SH(0,j) ;
                               db22 = w_det * lam_2mu * SH(1,j) ;
                               db23 = w_det * lam_el * SH(2,j) ;
                               db31 = w_det * lam_el * SH(0,j) ;
                               db32 = w_det * lam_el * SH(1,j) ;
                               db33 = w_det * lam_2mu * SH(2,j) ;
                               db41 = w_det * mu_el * SH(1,j) ;
                               db42 = w_det * mu_el * SH(0,j) ;
                               db51 = w_det * mu_el * SH(2,j) ;
                               db53 = w_det * mu_el * SH(0,j) ;
                               db62 = w_det * mu_el * SH(2,j) ;
                               db63 = w_det * mu_el * SH(1,j) ;
                               for(i=0;i<=j;i++)
                                  {
                                   it3 = 3*i ;
                                   ES(it3 , jt3)     +=
                                    SH(0,i)*db11 + SH(1,i)*db41 + SH(2,i)*db51 ;
                                   ES(it3+1 , jt3)   +=
                                      SH(1,i)*db21 + SH(0,i)*db41 ;
                                   ES(it3+2 , jt3)    +=
                                      SH(2,i)*db31 + SH(0,i)*db51 ;
                                   ES(it3 , jt3+1)   +=
                                      SH(0,i)*db12 + SH(1,i)*db42 ;
                                   ES(it3+1 , jt3+1) +=
                                    SH(1,i)*db22 + SH(0,i)*db42 + SH(2,i)*db62 ;
                                   ES(it3+2 , jt3+1)  +=  
                                      SH(2,i)*db32 + SH(1,i)*db62 ;
                                   ES(it3 , jt3+2)    += 
                                      SH(0,i)*db13 + SH(2,i)*db53 ;
                                   ES(it3+1 , jt3+2)  +=
                                      SH(1,i)*db23 + SH(2,i)*db63 ;
                                   ES(it3+2 , jt3+2) +=
                                    SH(2,i)*db33 + SH(0,i)*db53 + SH(1,i)*db63;
                                  }
                              }
                          }

                       else if( code == VISCO && nsd == 3)
                          {
                           for(j=0;j<nen;j++)
                              {
                               jt3 = 3*j ;
                               bcol[0]  = w_det * SH(0,j) ;
                               bcol[1]  = ZERO ;
                               bcol[2]  = ZERO ;
                               bcol[3]  = w_det * SH(1,j) ;
                               bcol[4]  = w_det * SH(2,j) ;
                               bcol[5]  = ZERO ;
                               full_back(dbar,bcol,el_diag,str_size) ;
                               db11 = bcol[0] ;
                               db21 = bcol[1] ;
                               db31 = bcol[2] ;
                               db41 = bcol[3] ;
                               db51 = bcol[4] ;
                               db61 = bcol[5] ;
                               bcol[0]  = ZERO ;
                               bcol[1]  = w_det * SH(1,j) ;
                               bcol[2]  = ZERO ;
                               bcol[3]  = w_det * SH(0,j) ;
                               bcol[4]  = ZERO ;
                               bcol[5]  = w_det * SH(2,j) ;
                               full_back(dbar,bcol,el_diag,str_size) ;
                               db12 = bcol[0] ;
                               db22 = bcol[1] ;
                               db32 = bcol[2] ;
                               db42 = bcol[3] ;
                               db52 = bcol[4] ;
                               db62 = bcol[5] ;
                               bcol[0]  = ZERO ;
                               bcol[1]  = ZERO ;
                               bcol[2]  = w_det * SH(2,j) ;
                               bcol[3]  = ZERO ;
                               bcol[4]  = w_det * SH(0,j) ;
                               bcol[5]  = w_det * SH(1,j) ;
                               full_back(dbar,bcol,el_diag,str_size) ;
                               db13 = bcol[0] ;
                               db23 = bcol[1] ;
                               db33 = bcol[2] ;
                               db43 = bcol[3] ;
                               db53 = bcol[4] ;
                               db63 = bcol[5] ;
                               for(i=0;i<=j;i++)
                                  {
                                   it3 = 3*i ;
                                   ES(it3 , jt3)     +=
                                    SH(0,i)*db11 + SH(1,i)*db41 + SH(2,i)*db51;
                                   ES(it3+1 , jt3)   +=
                                    SH(1,i)*db21 + SH(0,i)*db41 + SH(2,i)*db61;
                                   ES(it3+2 , jt3)    =
                                    SH(0,i)*db51 + SH(1,i)*db61 + SH(2,i)*db31;
                                   ES(it3 , jt3+1)   +=
                                    SH(0,i)*db12 + SH(1,i)*db42 + SH(2,i)*db52;
                                   ES(it3+1 , jt3+1) +=
                                    SH(1,i)*db22 + SH(0,i)*db42 + SH(2,i)*db62;
                                   ES(it3+2 , jt3+1)  =
                                    SH(0,i)*db52 + SH(1,i)*db62 + SH(2,i)*db32 ;
                                   ES(it3 , jt3+2)    =
                                    SH(0,i)*db13 + SH(1,i)*db43 + SH(2,i)*db53;
                                   ES(it3+1 , jt3+2)  =
                                    SH(1,i)*db23 + SH(0,i)*db43 + SH(2,i)*db63;
                                   ES(it3+2 , jt3+2) +=
                                    SH(0,i)*db53 + SH(1,i)*db63 + SH(2,i)*db33;
                                  }
                              }
                          }
                   break ;
                   
               }
                 
           }
  }  
/************************** end of lame_form *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   force_form   ***************
*/

/*  INTERFACE:   */

    force_form(
               int  code ,     /* code specifying elastic or visco */
               real  *rhs ,   /* temp array to hold element rhs contrib */
               ELEMENT_INFO  *info ,   /* ptr to element info struct */
               ELEMENT_MAT   *mat_pt ,  /* ptr to el material prop struct */
               ELEMENT_DATA  *el_pt    /* ptr to current element struct */
              )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine force_form
** force_form computes the "dumb" gravity contribution to the element
** right-hand side term for the elastic case, and the strain-rate
** term for the VE case.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      real    *grav , *shape_ptr , *det_ptr , *sh , *dbar , *wt ;
      real    w_det , btemp[6] ;
      int      i , j , ipt , active , jt2 , jt3 , nint , nen , nee ,
               ndof , str_size , dbar_size , save_shape , type , nsd , nsdp1 ;
               
      nsd = fe_sys.nsd ;
      nsdp1 = nsd + 1 ;
      ndof  = fe_sys.ndof ;
      nint = el_pt->nint ;
      grav  = mat_pt->bforce + ndof*(el_pt->mat) ;
      nen     = info->nen ;
      nee     = info->nee ;
      type = info->type ;
      str_size = info->stress_size ;
      dbar_size = info->dbar_size ;
      save_shape = fe_sys.save_shape;
      switch( nint )
         {
          case 4:
             wt = (type == BILIN) ? wt_2x2 : wt_tet ;
             break ;
          case 6:
             wt = wt_tri ;
             break ;
          case 8:
             wt = wt_2x2x2 ;
             break ;
          case 9:
             wt = wt_3x3 ;
             break ;
         }
      

      if( !(*(mat_pt->plastic + el_pt->mat)) )
         code = ELASTIC ;

      for (  i = 0 ; i < nee ; i++ )
          *(rhs + i) = ZERO ;
          

      if( code == VISCO )
         form_beta(el_pt  ,  mat_pt  , info) ;
          
      if( (code == ELASTIC) && time_data.elastic)
         {
          active = 0 ;
          for(i=0 ; i<ndof ; i++)
            if( *(grav + i) != ZERO )  active = 1 ;
          if( !active ) return ;
         }
        
      if( !save_shape )
         {
          shape_ptr = sh_temp ;
          det_ptr = det_temp ;
          shape( el_pt , info , shape_ptr , det_ptr , type , nen , nint , 1 ) ;
         }
      else
         {
          shape_ptr = el_pt -> shape ;
          det_ptr = el_pt -> deter ;
         }

      for (ipt = 0 ; ipt < nint ; ipt++)
           {
            sh = shape_ptr + ipt*nen*nsdp1 ;
            w_det = wt[ipt] * *(det_ptr + ipt) ;
            dbar = (el_pt -> dbar) + ipt*dbar_size ;
            
            switch( ndof )
               {
                case 1:
                   if( code == ELASTIC )
                      {
                       for(j=0;j<nen;j++)
                          {
                           *(rhs + j) += w_det * (*grav) * SH(nsd,j) ;
                          }
                      }
 
                    else if( code == VISCO )
                      {
                       move_real( (el_pt->bta)+ipt*str_size ,
                                   btemp , str_size ) ;
                       full_back(dbar,btemp,el_diag,str_size) ;

                       for(j=0;j<nen;j++)
                          {
                           *(rhs + j) += w_det *
                             (btemp[0] * SH(0,j) + btemp[1] * SH(1,j)) ;
                          }
                      }
                  break ;
                   
                case 2:
                   if( code == ELASTIC )
                      {
                       for(j=0;j<nen;j++)
                          {
                           jt2 = j+j ;
                           *(rhs + jt2)     +=
                              w_det * (*grav) * SH(nsd,j) ;
                           *(rhs + jt2 + 1) +=
                              w_det * (*(grav + 1)) * SH(nsd,j) ;
                          }
                      }
 
                    else if( code == VISCO )
                      {
                       move_real( (el_pt->bta)+ipt*str_size ,
                                   btemp , str_size ) ;
                       full_back(dbar,btemp,el_diag,str_size) ;

                       for(j=0;j<nen;j++)
                          {
                           jt2 = j+j ;

                           *(rhs + jt2)     += w_det *
                             (btemp[0] * SH(0,j) + btemp[3] * SH(1,j)) ;
                           *(rhs + jt2 + 1) += w_det *
                             (btemp[1] * SH(1,j) + btemp[3] * SH(0,j)) ;
                          }
                      }
                   break ;
                   
                case 3:
                   if( code == ELASTIC )
                      {
                       for(j=0;j<nen;j++)
                          {
                           jt3 = 3*j ;
                           *(rhs + jt3)     +=
                              w_det * (*grav) * SH(nsd,j) ;
                           *(rhs + jt3 + 1) +=
                              w_det * (*(grav + 1)) * SH(nsd,j) ;
                           *(rhs + jt3 + 2) +=
                              w_det * (*(grav + 2)) * SH(nsd,j) ;
                          }
                      }
 
                    else if( code == VISCO )
                      {
                       move_real( (el_pt->bta)+ipt*str_size ,
                                   btemp , str_size ) ;
                       full_back(dbar,btemp,el_diag,str_size) ;

                       if(nsd == 2)
                       {
                       for(j=0;j<nen;j++)
                          {
                           jt3 = 3*j ;

                           *(rhs + jt3)     += w_det * 
                             (btemp[0] * SH(0,j) + btemp[3] * SH(1,j)) ;
                           *(rhs + jt3 + 1) += w_det *
                             (btemp[1] * SH(1,j) + btemp[3] * SH(0,j)) ;
                           *(rhs + jt3 + 2) += w_det *
                             (btemp[4] * SH(0,j) + btemp[5] * SH(1,j)) ;
                          }
                       } else if (nsd == 3)
                       {
                       for(j=0;j<nen;j++)
                          {
                           jt3 = 3*j ;

                           *(rhs + jt3)     += w_det * (btemp[4] * SH(2,j) +
                             btemp[0] * SH(0,j) + btemp[3] * SH(1,j)) ;
                           *(rhs + jt3 + 1) += w_det * (btemp[5] * SH(2,j) +
                             btemp[1] * SH(1,j) + btemp[3] * SH(0,j)) ;
                           *(rhs + jt3 + 2) += w_det * (btemp[2] * SH(2,j) +
                             btemp[4] * SH(0,j) + btemp[5] * SH(1,j)) ;
                          }
                       }
                      }
                   break ;
                   
               }
                 
           }
  }  
/************************** end of force_form ********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   surf_form   ***************
*/

/*  INTERFACE:   */

     surf_form(
               int  code ,     /* code specifying elastic or visco */
               real  *rhs ,   /* temp array to hold element rhs contrib */
               ELEMENT_INFO  *info ,   /* ptr to element info struct */
               int  side ,    /* index number of the element side */
               real  *trac ,   /* array containg traction components */
               ELEMENT_DATA  *el_pt    /* ptr to current element struct */
              )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine surf_form
** surf_form computes the surface traction forcing term for the 
** right-hand-side in the finite element problem.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      int  i , nee , ndof , n1 , n2 , n3 , n4 , node1 , node2 , node3 , node4 , midnode, nsd , nsdp1 ;
      real tx , ty , tz , xx1 , yy1 , zz1 , xx2 , yy2 , zz2 , xx3 , yy3 , zz3 , 
            dx , dy , dl , ax , ay , az , bx , by , bz , cx , cy , cz , area;
     
      nsd = fe_sys.nsd ;
      nsdp1 = nsd + 1 ;
      nee     = info->nee ;
      ndof     = fe_sys.ndof ;
      
      for (  i = 0 ; i < nee ; i++ )
          *(rhs + i) = ZERO ;
          
      
      if(nsd == 2)
         {
/* ***** Applies surface tractions to an element edge ***** */
/* ***** This version applies two-point surface tractions to a
         bilinear element edge.  DO NOT USE SERENDIP OPTION ***** */
         node1 = side - 1 ;
         node2 = side ;
         if( node2 == 4 )  node2 = 0 ;
         n1 = *(el_pt->ien + node1) - 1 ;
         n2 = *(el_pt->ien + node2) - 1 ;
         xx1 = *(global.coords + n1*nsd) ;
         yy1 = *(global.coords + n1*nsd + 1) ;
         xx2 = *(global.coords + n2*nsd) ;
         yy2 = *(global.coords + n2*nsd + 1) ;
         dx = xx2 - xx1 ;
         dy = yy2 - yy1 ;
         dl = sqrt(dx*dx + dy*dy) ;

/* changed to global traction reference, rather than normal-tangential
   formulation, and changed to single midpoint - implemented 10/19/01  */
         switch( ndof )
            {
             case 1:
                *(rhs + node1) =
                  dl * PT5 * (*trac) ;

                *(rhs + node2) =
                  dl * PT5 * (*trac) ;
                break ;
                   
             case 2:
                *(rhs + 2*node1    ) =
                  dl * PT5 * (*trac) ;
                *(rhs + 2*node1 + 1) =
                  dl * PT5 * (*trac+1) ;

                *(rhs + 2*node2    ) =
                  dl * PT5 * (*trac) ;
                *(rhs + 2*node2 + 1) =
                  dl * PT5 * (*trac+1) ;
                break ;
                   
             case 3:
                *(rhs + 3*node1    ) =
                  dl * PT5 * (*trac) ;
                *(rhs + 3*node1 + 1) =
                  dl * PT5 * (*trac+1) ;
                *(rhs + 3*node1 + 2) =
                  dl * PT5 * (*trac+2) ;

                *(rhs + 3*node2    ) =
                  dl * PT5 * (*trac) ;
                *(rhs + 3*node2 + 1) =
                  dl * PT5 * (*trac+1) ;
                *(rhs + 3*node2 + 2) =
                  dl * PT5 * (*trac+2) ;
                break ;
         }
      } 
      else if (nsd == 3 && (info -> type == TET))
         {
/************************************************************************/
/*
This version implements one-point surface tractions at the face
centroid of tetrahedral elements.  
GL 10/23/98
*/
/************************************************************************/
         switch(side)
           {
            case 1:
              node1 = 2 ;
              node2 = 3 ;
              node3 = 4 ;
              break;
            case 2:
              node1 = 1 ;
              node2 = 3 ;
              node3 = 4 ;
              break;
            case 3:
              node1 = 1 ;
              node2 = 2 ;
              node3 = 4 ;
              break;
            case 4:
              node1 = 1 ;
              node2 = 2 ;
              node3 = 3 ;
              break;
         }
         n1 = *(el_pt->ien + node1) - 1 ;  /* look up global node #'s */
         n2 = *(el_pt->ien + node2) - 1 ;
         n3 = *(el_pt->ien + node3) - 1 ;
         xx1 = *(global.coords + n1*nsd) ;     /* then get the global coords */
         yy1 = *(global.coords + n1*nsd + 1) ;
         zz1 = *(global.coords + n1*nsd + 2) ;
         xx2 = *(global.coords + n2*nsd) ;
         yy2 = *(global.coords + n2*nsd + 1) ;
         zz2 = *(global.coords + n2*nsd + 2) ;
         xx3 = *(global.coords + n3*nsd) ;
         yy3 = *(global.coords + n3*nsd + 1) ;
         zz3 = *(global.coords + n3*nsd + 2) ;

         ax = xx2 - xx1 ;    /* compute the edge vectors... */
         ay = yy2 - yy1 ;
         az = zz2 - zz1 ;
         bx = xx3 - xx1 ;
         by = yy3 - yy1 ;
         bz = zz3 - zz1 ;

         cx = ay*bz - by*az ;   /* ... and compute their cross product...*/
         cy = az*bx - bz*ax ;
         cz = ax*by - bx*ay ;

         area = sqrt(cx*cx + cy*cy + cz*cz) / 6.0 ;  /* ... and get the area
(divided    by three) */

                *(rhs + 3*node1    ) =  area * (*trac) ;   
                *(rhs + 3*node1 + 1) =  area * (*(trac+1)) ;
                *(rhs + 3*node1 + 2) =  area * (*(trac+2)) ;

                *(rhs + 3*node2    ) =  area * (*trac) ;
                *(rhs + 3*node2 + 1) =  area * (*(trac+1)) ;
                *(rhs + 3*node2 + 2) =  area * (*(trac+2)) ;

                *(rhs + 3*node3    ) =  area * (*trac) ;
                *(rhs + 3*node3 + 1) =  area * (*(trac+1)) ;
                *(rhs + 3*node3 + 2) =  area * (*(trac+2)) ;
           }

      else if (nsd == 3 && (info -> type == TRILIN))
         {
/************************************************************************/
/*
This version implements one-point surface tractions at the face
centroid of brick elements.  
GL 10/19/01
*/
/************************************************************************/
         switch(side)
           {
            case 1:
              node1 = 1 ;
              node2 = 2 ;
              node3 = 3 ;
              node4 = 4 ;
              break;
            case 2:
              node1 = 1 ;
              node2 = 2 ;
              node3 = 6 ;
              node4 = 5 ;
              break;
            case 3:
              node1 = 2 ;
              node2 = 3 ;
              node3 = 7 ;
              node4 = 6 ;
              break;
            case 4:
              node1 = 3 ;
              node2 = 4 ;
              node3 = 8 ;
              node4 = 7 ;
              break;
            case 5:
              node1 = 4 ;
              node2 = 1 ;
              node3 = 5 ;
              node4 = 8 ;
              break;
            case 6:
              node1 = 5 ;
              node2 = 6 ;
              node3 = 7 ;
              node4 = 8 ;
              break;
         }
         n1 = *(el_pt->ien + node1) - 1 ;  /* look up global node #'s */
         n2 = *(el_pt->ien + node2) - 1 ;
         n3 = *(el_pt->ien + node3) - 1 ;
         n4 = *(el_pt->ien + node4) - 1 ;
         xx1 = *(global.coords + n1*nsd) ;     /* then get the global coords */
         yy1 = *(global.coords + n1*nsd + 1) ;
         zz1 = *(global.coords + n1*nsd + 2) ;
         xx2 = *(global.coords + n2*nsd) ;
         yy2 = *(global.coords + n2*nsd + 1) ;
         zz2 = *(global.coords + n2*nsd + 2) ;
         xx3 = *(global.coords + n3*nsd) ;
         yy3 = *(global.coords + n3*nsd + 1) ;
         zz3 = *(global.coords + n3*nsd + 2) ;

         ax = xx2 - xx1 ;    /* compute the edge vectors... */
         ay = yy2 - yy1 ;
         az = zz2 - zz1 ;
         bx = xx3 - xx1 ;
         by = yy3 - yy1 ;
         bz = zz3 - zz1 ;

         cx = ay*bz - by*az ;   /* ... and compute their cross product...*/
         cy = az*bx - bz*ax ;
         cz = ax*by - bx*ay ;

         area = sqrt(cx*cx + cy*cy + cz*cz) / 6.0 ;  /* ... and get the area
(divided    by three) */

                *(rhs + 3*node1    ) =  area * (*trac) ;   
                *(rhs + 3*node1 + 1) =  area * (*(trac+1)) ;
                *(rhs + 3*node1 + 2) =  area * (*(trac+2)) ;

                *(rhs + 3*node2    ) =  area * (*trac) ;
                *(rhs + 3*node2 + 1) =  area * (*(trac+1)) ;
                *(rhs + 3*node2 + 2) =  area * (*(trac+2)) ;

                *(rhs + 3*node3    ) =  area * (*trac) ;
                *(rhs + 3*node3 + 1) =  area * (*(trac+1)) ;
                *(rhs + 3*node3 + 2) =  area * (*(trac+2)) ;
           }

/*--------------------------------------------------------------------*/
/* **** This block of code is obsolete, but kept for future reference ****
      switch( info -> type )
         {
          case BILIN:
             switch( ndof )
                {
                 case 1:
                    *(rhs + node1) = tz * PT5 ;
                    *(rhs + node2) = tz * PT5 ;
                    break ;
                 
                 case 2:
                    *(rhs + 2*node1    ) = tx * PT5 ;
                    *(rhs + 2*node1 + 1) = ty * PT5 ;
                    *(rhs + 2*node2    ) = tx * PT5 ;
                    *(rhs + 2*node2 + 1) = ty * PT5 ;
                    break ;
                 
                 case 3:
                    *(rhs + 3*node1    ) = tx * PT5 ;
                    *(rhs + 3*node1 + 1) = ty * PT5 ;
                    *(rhs + 3*node1 + 2) = tz * PT5 ;
                    *(rhs + 3*node2    ) = tx * PT5 ;
                    *(rhs + 3*node2 + 1) = ty * PT5 ;
                    *(rhs + 3*node2 + 2) = tz * PT5 ;
                    break ;
                }
             break ;

          case SERENDIP:
             midnode = node1 + 4 ;
             switch( ndof )
                {
                 case 1:
                    *(rhs + node1)   = tz / SIX ;
                    *(rhs + node2)   = tz / SIX ;
                    *(rhs + midnode) = tz * TWO / THREE ;
                    break ;
                    
                 case 2:
                    *(rhs + 2*node1    )   = tx / SIX ;
                    *(rhs + 2*node1 + 1)   = ty / SIX ;
                    *(rhs + 2*node2    )   = tx / SIX ;
                    *(rhs + 2*node2 + 1)   = ty / SIX ;
                    *(rhs + 2*midnode    ) = tx * TWO / THREE ;
                    *(rhs + 2*midnode + 1) = ty * TWO / THREE ;
                    break ;
                    
                 case 3:
                    *(rhs + 3*node1    )   = tx / SIX ;
                    *(rhs + 3*node1 + 1)   = ty / SIX ;
                    *(rhs + 3*node1 + 2)   = tz / SIX ;
                    *(rhs + 3*node2    )   = tx / SIX ;
                    *(rhs + 3*node2 + 1)   = ty / SIX ;
                    *(rhs + 3*node2 + 2)   = tz / SIX ;
                    *(rhs + 3*midnode    ) = tx * TWO / THREE ;
                    *(rhs + 3*midnode + 1) = ty * TWO / THREE ;
                    *(rhs + 3*midnode + 2) = tz * TWO / THREE ;
                    break ;
                }
             break ;
         }
*************************************************** */
/*--------------------------------------------------------------------*/

  }  
/************************** end of surf_form *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   shape   ***************
*/

/*  INTERFACE:   */

     shape(
           ELEMENT_DATA  *el_pt  ,  /* ptr to current element struct */
           ELEMENT_INFO  *info ,   /* ptr to element info struct */
           real  *sh_pt ,   /* el storage to hold shape functions */
           real  *det_pt ,   /* el storage to hold determinant */
           int  type ,    /* element type number */
           int  nen ,    /* number of element nodes */
           int  nint ,   /* number of Gauss integration points */
           int  all     /* flag indicating we need derivatives also */
          )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine shape
** shape computes the shape functions and gradients for a single element.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    real  xs[3][3] , hold,hold1,hold2 ;
    real  *sh , *det ;
    int    i , j , ipt , rule , nsd , nsdp1 ;
    
    nsd = fe_sys.nsd ;
    nsdp1 = nsd + 1 ;
    for(ipt=0 ; ipt<nint ; ipt++)
       {
        sh = sh_pt + ipt*nen*nsdp1 ;
        det = det_pt + ipt ;
        
        switch( type )
           {
            case TET:
                  for(i=0 ; i<nen ; i++)
                     {
                      if( all )
                         {
                          *(sh + i*nsdp1) = PSH_TET(0,i,ipt) ;
                          *(sh + i*nsdp1 + 1) = PSH_TET(1,i,ipt) ;
                          *(sh + i*nsdp1 + 2) = PSH_TET(2,i,ipt) ;
                         }
                      *(sh + i*nsdp1 + nsd) = PSH_TET(nsd,i,ipt) ;
                     }
    
               break ;

            case BILIN:
               for(i=0 ; i<nen ; i++)
                  {
                   if( all )
                      {
                       *(sh + i*nsdp1) = PSH_BILIN(0,i,ipt) ;
                       *(sh + i*nsdp1 + 1) = PSH_BILIN(1,i,ipt) ;
                      }
                   *(sh + i*nsdp1 + nsd) = PSH_BILIN(nsd,i,ipt) ;
                  }
               if(el_pt->degen)
                  {
                   for(j=0 ; j<nsdp1 ; j++)
                      {
                       *(sh + 2*nsdp1 + j) += *(sh + 3*nsdp1 + j) ;
                       *(sh + 3*nsdp1 + j) = ZERO ;
                      }
                  }
    
               break ;
    
            case TRILIN:
               for(i=0 ; i<nen ; i++)
                  {
                   if(nint == 6)
                      {
					   if( all )
						  {
						   *(sh + i*nsdp1) = PSH_TRILIN6(0,i,ipt) ;
						   *(sh + i*nsdp1 + 1) = PSH_TRILIN6(1,i,ipt) ;
						   *(sh + i*nsdp1 + 2) = PSH_TRILIN6(2,i,ipt) ;
						  }
					   *(sh + i*nsdp1 + nsd) = PSH_TRILIN6(nsd,i,ipt) ;
                      }
                   else if(nint == 8)
                      {
					   if( all )
						  {
						   *(sh + i*nsdp1) = PSH_TRILIN8(0,i,ipt) ;
						   *(sh + i*nsdp1 + 1) = PSH_TRILIN8(1,i,ipt) ;
						   *(sh + i*nsdp1 + 2) = PSH_TRILIN8(2,i,ipt) ;
						  }
					   *(sh + i*nsdp1 + nsd) = PSH_TRILIN8(nsd,i,ipt) ;
                      }
                  }
               if(el_pt->degen == 1)  /* triangular wedge */
                  {
                   for(j=0 ; j<nsdp1 ; j++)
                      {
                       *(sh + 2*nsdp1 + j) += *(sh + 3*nsdp1 + j) ;
                       *(sh + 3*nsdp1 + j) = ZERO ;
                       *(sh + 6*nsdp1 + j) += *(sh + 7*nsdp1 + j) ;
                       *(sh + 7*nsdp1 + j) = ZERO ;
                      }
                  }
               if(el_pt->degen == 2)  /* tetrahedron */
                  {
                   for(j=0 ; j<nsdp1 ; j++)
                      {
                       *(sh + 2*nsdp1 + j) += *(sh + 3*nsdp1 + j) ;
                       *(sh + 3*nsdp1 + j) = ZERO ;
                       *(sh + 4*nsdp1 + j) += 
                         (*(sh + 5*nsdp1 + j) + *(sh + 6*nsdp1 + j) + *(sh + 7*nsdp1 + j)) ;
                       *(sh + 5*nsdp1 + j) = ZERO ;
                       *(sh + 6*nsdp1 + j) = ZERO ;
                       *(sh + 7*nsdp1 + j) = ZERO ;
                      }
                  }
    
               break ;
    
            case SERENDIP:
               if(nint == 4) rule = 0 ;
               else if(nint == 9) rule = 1 ;
               for(i=0 ; i<nen ; i++)
                  {
                   if( all )
                      {
                       *(sh + i*nsdp1)       = PSH_SEREN(0,i,ipt,rule) ;
                       *(sh + i*nsdp1 + 1)   = PSH_SEREN(1,i,ipt,rule) ;
                      }
                   *(sh + i*nsdp1 + nsd) = PSH_SEREN(nsd,i,ipt,rule) ;
                  }
               if(el_pt->degen)
                  {
                   for(j=0 ; j<nsdp1 ; j++)
                      {
                       *(sh + 2*nsdp1 + j) += *(sh + 3*nsdp1 + j) +
                                              *(sh + 6*nsdp1 + j) ;
                       *(sh + 3*nsdp1 + j) = ZERO ;
                       *(sh + 6*nsdp1 + j) = ZERO ;
                      }
                  }
    
               break ;
           }
    
          if( !all )  return ;
          
          for(i=0 ; i<nsd ; i++)
             {
              for(j=0 ; j<nsd ; j++)
/* note that the first argument counts the physical space coordinate x, 
while the second counts the parent space coordinate Tse. 
Hence xs[j][i] = dX(j) / dTse(i).
*/
                xs[j][i] = dotsh( global.coords , i , j ,
                                        sh , el_pt , info , nen , nsd , 0 ) ;
             }
    
          if(nsd == 2){

             *det = xs[0][0]*xs[1][1] - xs[0][1]*xs[1][0] ;

          } else if (nsd==3) {

             *det = xs[0][0]*xs[1][1]*xs[2][2]
                   -xs[2][0]*xs[1][1]*xs[0][2]
                   +xs[0][1]*xs[1][2]*xs[2][0]
                   -xs[2][1]*xs[1][2]*xs[0][0]
                   +xs[0][2]*xs[1][0]*xs[2][1]
                   -xs[2][2]*xs[1][0]*xs[0][1];

          } else {
            attempt = DET_ERR;
            printf("NSD %d Not == 2 or 3!\n",nsd);
          }


          if( *det <= ZERO )
             {
              attempt = DET_ERR ;
              printf("Bad determinant in element %d: %8.5g\n",
                      el_pt->nel,*det) ;

             }
/* jwp cutting out the determinant division: it doesn't work in 3d.
hence alter the 2d case below as well.
          else
             {
              for(i=0 ; i<nsd ; i++)
                 {
                  for(j=0 ; j<nsd ; j++)
                    xs[i][j] /= (*det) ;
                 }
             }
*/
/* 
C   compute gradient of shape fns w/respect to x
C                                           /     \ -1
C      dNj       dNj       dTse     dNj    |  dX   |
C     -----  =  -----  *  ------ = ----- * | ----- |
C      dX        dTse      dX       dTse   |  dTse |
C                                           \     /
C
*/
/* this code corrected as per JWP bug fix of Sept. 2000 */

          if(nsd == 2){

          for(i=0 ; i<nen ; i++)
             {
              hold = (xs[1][1]* *(sh + i*nsdp1) -
                     xs[1][0]* *(sh + i*nsdp1 + 1)) ;

              hold1 = (xs[0][0]* *(sh + i*nsdp1 + 1) -
                   xs[0][1]* *(sh + i*nsdp1)) ;

              *(sh + i*nsdp1) = hold/ *det ;
              *(sh + i*nsdp1 + 1) = hold1/ *det ;
             }

          } else if (nsd == 3){

          for(i=0 ; i<nen ; i++)
             {
              hold = (xs[1][1]*xs[2][2] - xs[1][2]*xs[2][1]) * *(sh + i*nsdp1) +
                ( xs[1][2]*xs[2][0] - xs[1][0]*xs[2][2]) * *(sh + i*nsdp1 + 1) +
                ( xs[1][0]*xs[2][1] - xs[1][1]*xs[2][0]) * *(sh + i*nsdp1 + 2) ;

              hold1= (xs[0][2]*xs[2][1] - xs[0][1]*xs[2][2]) * *(sh + i*nsdp1) +
                ( xs[0][0]*xs[2][2] - xs[0][2]*xs[2][0]) * *(sh + i*nsdp1 + 1) +
                ( xs[0][1]*xs[2][0] - xs[0][0]*xs[2][1]) * *(sh + i*nsdp1 + 2) ;

              hold2= (xs[0][1]*xs[1][2] - xs[0][2]*xs[1][1]) * *(sh + i*nsdp1) +
                ( xs[0][2]*xs[1][0] - xs[0][0]*xs[1][2]) * *(sh + i*nsdp1 + 1) +
                ( xs[0][0]*xs[1][1] - xs[0][1]*xs[1][0]) * *(sh + i*nsdp1 + 2) ;

              *(sh + i*nsdp1) = hold/ *det ;
              *(sh + i*nsdp1 + 1) = hold1/ *det ;
              *(sh + i*nsdp1 + 2) = hold2/ *det ;
             }

          } else {
            attempt = DET_ERR;
            printf("NSD %d Not == 2 or 3!\n",nsd);
          }
       }
   }
/************************** end of shape *************************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   dotsh   ***************
*/

/*  INTERFACE:   */

real dotsh(
           real  *array ,  /* array of nodal quantities to interpolate */
           int  i ,     /* derivative component index number */
           int  j ,     /* dof index number of quantity component */
           real  *sh ,   /* el storage containing shape functions */
           ELEMENT_DATA  *el_pt  ,  /* ptr to current element struct */
           ELEMENT_INFO  *info ,   /* ptr to element info struct */
           int  nen ,    /* number of element nodes */
           int  ndim ,    /* number of dimensions in array */
           int  st_flag      /* flag to invoke split node adjustment */
          )

/*  RETURN VALUE:  the element-interpolated sum value */
/*  DESCRIPTION:   */
/*
** Function dotsh
** dotsh computes the dot product of a set of shape-functions (or gradients)
** with a vector of values, hence computing an interpolated coordinate
** or function. If st_flag, x is assumed to represent displacement, and
** a correction for split-nodes is applied prior to the dot product.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    real  sum = ZERO ;
    real  x ;
    int    n , node , nsd , nsdp1 ;
    
    nsd = fe_sys.nsd ;
    nsdp1 = nsd + 1 ;
    for(n=0 ; n<nen ; n++)
        {
         node = *(el_pt->ien + n) - 1 ;
         x = *(array + node*ndim + j) ;

         if( st_flag && time_data.do_slip )
            x += adfldp( j , el_pt->nel , node , info ) ;
         
         sum += x * SH(i,n) ;
        }
    return( sum ) ;
   }
/************************** end of dotsh *************************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   adfldp   ***************
*/

/*  INTERFACE:   */

real adfldp(
            int  j ,     /* dof index number of slip component */
            int  nel ,    /* number of current element */
            int  node ,    /* number of current node */
            ELEMENT_INFO  *info    /* ptr to element info struct */
           )

/*  RETURN VALUE:  the required additional nodal displacement */
/*  DESCRIPTION:   */
/*
** Function adfldp
** adfldp returns to dotsh() the added nodal displacements
** needed to account for split node fault slip in calculating strain
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    int    i , ndof , nfnode ;
    int    *slip_list ;
    real   *slip_val ;
    
    ndof     = fe_sys.ndof ;
    nfnode = info -> nfnode ;
    slip_list = info -> slip_list ;
    slip_val = info -> slip_val ;
    for(i=0 ; i<nfnode ; i++)
        {
         if( *(slip_list + 2*i) == (nel-1) && 
             *(slip_list + 2*i +1) == (node+1) )
            return( *(slip_val + ndof*i + j) ) ;
        }
    return( ZERO ) ;
   }
/************************** end of adfldp ************************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   p_shape   ***************
*/

/*  INTERFACE:   */

     p_shape(
             GROUP   *grp_ptr     /* pointer to current element group */
            )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine p_shape
** p_shape computes the parent-space shape function for common element types.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

   {
    int  ipt , rule , nen , nint , nel , numel , type , nsd , nsdp1 , n ;
    real  xpt , ypt, zpt ;
    ELEMENT_INFO  *info ;
    ELEMENT_DATA   *el_pt ;
    
/*  printf("Doing parent functions...\n") ;  */

    nsd = fe_sys.nsd ;
    nsdp1 = nsd + 1 ;
    switch((grp_ptr->el_info)->type)
       {
        case QUAKE:
           return ; /* no shape functions for quake elements */
           break ;

        case TET:
           for(ipt = 0; ipt < 4; ipt++)
              {
               xpt = x_tet[ipt] ;
               ypt = y_tet[ipt] ;
               zpt = z_tet[ipt] ;

/*
C
C  3D parent coordinates : (r,s,t)
C     define u = 1 - r - s - t
C  Shape functions and derivatives :
C
C      |               |  pN  |  pN  |  pN              s
C      |       N       | ---- | ---- | ----             |
C      |               |  pr  |  ps  |  pt              3
C -----|---------------|------|------|------            |
C   1  | 1 - r - s - t |  -1  |  -1  |  -1
C -----|---------------|------|------|------            |
C   2  |       r       |  1   |  0   |  0               |
C -----|---------------|------|------|------            |
C   3  |       s       |  0   |  1   |  0               1-------- 2 --r
C -----|---------------|------|------|------           /
C   4  |       t       |  0   |  0   |  1            /
C -----|---------------|------|------|------       4
C                                                /
C                                              t
*/
               PSH_TET(0,0,ipt) = -ONE ;
               PSH_TET(1,0,ipt) = -ONE ;
               PSH_TET(2,0,ipt) = -ONE ;
               PSH_TET(nsd,0,ipt) = ONE - xpt - ypt - zpt;

               PSH_TET(0,1,ipt) = ONE ;
               PSH_TET(1,1,ipt) = ZERO ;
               PSH_TET(2,1,ipt) = ZERO ;
               PSH_TET(nsd,1,ipt) = xpt ;

               PSH_TET(0,2,ipt) = ZERO  ;
               PSH_TET(1,2,ipt) = ONE;
               PSH_TET(2,2,ipt) = ZERO ;
               PSH_TET(nsd,2,ipt) = ypt ;

               PSH_TET(0,3,ipt) = ZERO  ;
               PSH_TET(1,3,ipt) = ZERO  ;
               PSH_TET(2,3,ipt) = ONE;
               PSH_TET(nsd,3,ipt) = zpt ;
              }

           break ;

        case BILIN:
           for(ipt=0 ; ipt<4 ; ipt++)
              {
               xpt = x_2x2[ipt] ;
               ypt = y_2x2[ipt] ;
               xpt /= (real) sqrt(THREE) ;
               ypt /= (real) sqrt(THREE) ;
               
               PSH_BILIN(0,0,ipt) = PT25 * (ypt - ONE) ;
               PSH_BILIN(1,0,ipt) = PT25 * (xpt - ONE) ; 
               PSH_BILIN(nsd,0,ipt) = PT25 * (ONE - xpt) * (ONE - ypt) ;

               PSH_BILIN(0,1,ipt) = PT25 * (ONE - ypt) ;
               PSH_BILIN(1,1,ipt) = (-PT25) * (ONE + xpt) ; 
               PSH_BILIN(nsd,1,ipt) = PT25 * (ONE + xpt) * (ONE - ypt) ;

               PSH_BILIN(0,2,ipt) = PT25 * (ONE + ypt) ;
               PSH_BILIN(1,2,ipt) = PT25 * (ONE + xpt) ; 
               PSH_BILIN(nsd,2,ipt) = PT25 * (ONE + xpt) * (ONE + ypt) ;

               PSH_BILIN(0,3,ipt) = (-PT25) * (ONE + ypt) ;
               PSH_BILIN(1,3,ipt) = PT25 * (ONE - xpt) ; 
               PSH_BILIN(nsd,3,ipt) = PT25 * (ONE - xpt) * (ONE + ypt) ;
              }

           break ;

        case TRILIN:
             for(ipt=0 ; ipt<6 ; ipt++)
              {
               for(n=0 ; n<8 ; n++)
                  {
                   PSH_TRILIN6(0,n,ipt) = ra[n]*(0.5+sa[n]*y_tri[ipt])*(0.5+ta[n]*z_tri[ipt]);
                   PSH_TRILIN6(1,n,ipt) = sa[n]*(0.5+ra[n]*x_tri[ipt])*(0.5+ta[n]*z_tri[ipt]);
                   PSH_TRILIN6(2,n,ipt) = ta[n]*(0.5+ra[n]*x_tri[ipt])*(0.5+sa[n]*y_tri[ipt]);
                   PSH_TRILIN6(nsd,n,ipt) =
                     (0.5+ra[n]*x_tri[ipt])*(0.5+sa[n]*y_tri[ipt])*(0.5+ta[n]*z_tri[ipt]);
                  }
              }

             for(ipt=0 ; ipt<8 ; ipt++)
              {
               xpt = x_2x2x2[ipt] ;
               ypt = y_2x2x2[ipt] ;
               zpt = z_2x2x2[ipt] ;
               xpt /= (real) sqrt(THREE) ;
               ypt /= (real) sqrt(THREE) ;
               zpt /= (real) sqrt(THREE) ;
               
               for(n=0 ; n<8 ; n++)
                  {
                   PSH_TRILIN8(0,n,ipt) = ra[n]*(0.5+sa[n]*ypt)*(0.5+ta[n]*zpt);
                   PSH_TRILIN8(1,n,ipt) = sa[n]*(0.5+ra[n]*xpt)*(0.5+ta[n]*zpt);
                   PSH_TRILIN8(2,n,ipt) = ta[n]*(0.5+ra[n]*xpt)*(0.5+sa[n]*ypt);
                   PSH_TRILIN8(nsd,n,ipt) =
                     (0.5+ra[n]*xpt)*(0.5+sa[n]*ypt)*(0.5+ta[n]*zpt);
                  }
              }

           break ;

        case SERENDIP:
           for(rule=0 ; rule<2 ; rule++)
               {
                for(ipt=0 ; ipt<9 ; ipt++)
                  {
                   if(rule == 0 && ipt >= 4) continue ;
                   
                   if(rule == 0)
                      {
                       xpt = x_2x2[ipt] ;
                       ypt = y_2x2[ipt] ;
                       xpt /= (real) sqrt(THREE) ;
                       ypt /= (real) sqrt(THREE) ;
                      }
                   else if(rule == 1)
                      {
                       xpt = x_3x3[ipt] ;
                       ypt = y_3x3[ipt] ;
                       xpt *= (real) sqrt(THREE/FIVE) ;
                       ypt *= (real) sqrt(THREE/FIVE) ;
                      }
               
                   PSH_SEREN(0,4,ipt,rule) = xpt * (ypt - ONE) ;
                   PSH_SEREN(1,4,ipt,rule) = PT5 * (xpt*xpt - ONE) ;
                   PSH_SEREN(nsd,4,ipt,rule) = PT5 * (ONE - xpt*xpt) *
                                               (ONE - ypt) ;
               
                   PSH_SEREN(0,5,ipt,rule) = PT5 * (ONE - ypt*ypt) ;
                   PSH_SEREN(1,5,ipt,rule) = (-ypt) * (ONE + xpt) ;
                   PSH_SEREN(nsd,5,ipt,rule) = PT5 * (ONE + xpt) *
                                               (ONE - ypt*ypt) ;
               
                   PSH_SEREN(0,6,ipt,rule) = (-xpt) * (ONE + ypt) ;
                   PSH_SEREN(1,6,ipt,rule) = PT5 * (ONE - xpt*xpt) ;
                   PSH_SEREN(nsd,6,ipt,rule) = PT5 * (ONE - xpt*xpt) *
                                               (ONE + ypt) ;
               
                   PSH_SEREN(0,7,ipt,rule) = PT5 * (ypt*ypt - ONE) ;
                   PSH_SEREN(1,7,ipt,rule) = ypt * (xpt - ONE) ;
                   PSH_SEREN(nsd,7,ipt,rule) = PT5 * (ONE - xpt) *
                                               (ONE - ypt*ypt) ;
               
                   PSH_SEREN(0,0,ipt,rule) = PT25 * (ypt - ONE) -
                         PT5 * (PSH_SEREN(0,7,ipt,rule) +
                         PSH_SEREN(0,4,ipt,rule)) ;
                   PSH_SEREN(1,0,ipt,rule) = PT25 * (xpt - ONE) -
                         PT5 * (PSH_SEREN(1,7,ipt,rule) +
                         PSH_SEREN(1,4,ipt,rule)) ;
                   PSH_SEREN(nsd,0,ipt,rule) = PT25 * (ONE - xpt) *
                                               (ONE - ypt) -
                         PT5 * (PSH_SEREN(nsd,7,ipt,rule) +
                         PSH_SEREN(nsd,4,ipt,rule)) ;
               
                   PSH_SEREN(0,1,ipt,rule) = PT25 * (ONE - ypt) -
                         PT5 * (PSH_SEREN(0,4,ipt,rule) +
                         PSH_SEREN(0,5,ipt,rule)) ;
                   PSH_SEREN(1,1,ipt,rule) = (-PT25) * (ONE + xpt) -
                         PT5 * (PSH_SEREN(1,4,ipt,rule) +
                         PSH_SEREN(1,5,ipt,rule)) ;
                   PSH_SEREN(nsd,1,ipt,rule) = PT25 * (ONE + xpt) *
                                               (ONE - ypt) -
                         PT5 * (PSH_SEREN(nsd,4,ipt,rule) +
                         PSH_SEREN(nsd,5,ipt,rule)) ;
               
                   PSH_SEREN(0,2,ipt,rule) = PT25 * (ONE + ypt) -
                         PT5 * (PSH_SEREN(0,5,ipt,rule) +
                         PSH_SEREN(0,6,ipt,rule)) ;
                   PSH_SEREN(1,2,ipt,rule) = PT25 * (ONE + xpt) -
                         PT5 * (PSH_SEREN(1,5,ipt,rule) +
                         PSH_SEREN(1,6,ipt,rule)) ;
                   PSH_SEREN(nsd,2,ipt,rule) = PT25 * (ONE + xpt) *
                                               (ONE + ypt) -
                         PT5 * (PSH_SEREN(nsd,5,ipt,rule) +
                         PSH_SEREN(nsd,6,ipt,rule)) ;
               
                   PSH_SEREN(0,3,ipt,rule) = (-PT25) * (ONE + ypt) -
                         PT5 * (PSH_SEREN(0,6,ipt,rule) +
                         PSH_SEREN(0,7,ipt,rule)) ;
                   PSH_SEREN(1,3,ipt,rule) = PT25 * (ONE - xpt) -
                         PT5 * (PSH_SEREN(1,6,ipt,rule) +
                         PSH_SEREN(1,7,ipt,rule)) ;
                   PSH_SEREN(nsd,3,ipt,rule) = PT25 * (ONE - xpt) *
                                               (ONE + ypt) -
                         PT5 * (PSH_SEREN(nsd,6,ipt,rule) +
                         PSH_SEREN(nsd,7,ipt,rule)) ;
               
                  }
               }

           break ;
       }

    if( fe_sys.save_shape )
       {
        info = grp_ptr->el_info ;
        nen     = info->nen ;
        numel     = info->numel ;
        type      = info->type ;

/*  printf("Allocating memory for shape functions...\n") ;  */

        for( nel = 0 ; nel < numel ; nel++)
            {
             el_pt = grp_ptr->el_data + nel ;
             nint = el_pt->nint ;
        
             el_pt->shape =
                  (real *) calloc(nsdp1*nen*nint,sizeof(real)) ;
             el_pt->deter = (real *) calloc(nint,sizeof(real)) ;

             if(el_pt->shape == NULL || el_pt->deter == NULL)
                attempt = SH_MEM ;
            }
        completion() ;

/*  printf("Doing daughter functions...\n") ;  */

        for( nel = 0 ; nel < numel ; nel++)
            {
             el_pt = grp_ptr->el_data + nel ;
             nint = el_pt->nint ;

             shape( el_pt , grp_ptr->el_info , el_pt->shape , el_pt->deter , type ,
                    nen , nint , 1 ) ;
            }
       }
       
   }
/************************** end of p_shape ***********************************/ 

 
/*
** Routine form_smooth
** form_smooth modifies element right-hand-sides for pressure smoothing.
**/
/*  - disabled -
form_smooth(grp_ptr , ypflag )
GROUP   *grp_ptr ;
int  ypflag ;
     {
      int     numel , nint , nel , i , j , ipt , nsd , nsdp1 ,
              ndof , nen , str_size , save_shape ;
      real  w_det , pressure ;
      real  *shape_ptr , *det_ptr , *stress , *wt ;

      ELEMENT_DATA   *el_pt ;


      nsd = fe_sys.nsd ;
      nsdp1 = nsd + 1 ;
      ndof = fe_sys.ndof ;
      if( ndof == 1 ) return ;
      numel = (grp_ptr->el_info)->numel ;
      nen = (grp_ptr->el_info)->nen ;
      str_size = (grp_ptr->el_info)->stress_size ;
      save_shape = fe_sys.save_shape;
      
      switch( nint )
         {
          case 4:
             wt = (nsd == 2) ? wt_2x2 : wt_tet;
             break ;
          case 9:
             wt = wt_3x3 ;
             break ;

         }

      switch((grp_ptr->el_info)->type)
          {
           case TET:
           case BILIN:
           case SERENDIP:
               for ( nel = 0 ; nel < numel ; nel++)
                   {
                    el_pt = grp_ptr->el_data + nel ;
                    nint = el_pt->nint ;
                    if( !save_shape )
                       {
                        shape_ptr = sh_temp ;
                        det_ptr = det_temp ;
                        shape( el_pt , grp_ptr->el_info , shape_ptr , det_ptr ,
                              (grp_ptr->el_info)->type ,
                              (grp_ptr->el_info)->nen , nint , 1 ) ;
                       }
                    else
                       {
                        shape_ptr = el_pt -> shape ;
                        det_ptr = el_pt -> deter ;
                       }

                    for (  i = 0 ; i < nen ; i++ )
                       *(el_rhs + i) = ZERO ;

                    pressure = ZERO ;
                    for (ipt = 0 ; ipt < nint ; ipt++)
                       {
                        stress = (el_pt->stress) + ipt*str_size ;
                        switch( ndof )
                          {
                           case 2:
                              pressure += ( (*stress + *(stress + 1)) /
                                             (TWO * nint) ) ;
                              break ;

                           case 3:
                              pressure += ( (*stress + *(stress + 1) +
                                          *(stress + 2)) / (THREE * nint) ) ;
                              break ;
                          }
                       }

                    for (ipt = 0 ; ipt < nint ; ipt++)
                       {
                        w_det = wt[ipt] * *(det_ptr + ipt) ;
                        for(j=0;j<nen;j++)
                           {
                            *(el_rhs+j) += w_det*pressure ;
                           }
                       }

                    addsmooth( global.pnode , el_rhs ,
                              grp_ptr->el_info , el_pt ) ;

                    if( ypflag == 1 )
                       {
                        for (  i = 0 ; i < nen ; i++ )
                           *(el_rhs + i) = ZERO ;

                        for (ipt = 0 ; ipt < nint ; ipt++)
                           {
                            w_det = wt[ipt] * *(det_ptr + ipt) ;
                            for(j=0;j<nen;j++)
                               {
                                *(el_rhs+j) += w_det ;
                               }
                           }

                        addsmooth( global.yp , el_rhs ,
                                  grp_ptr->el_info , el_pt ) ;
                       }
                   }
               break ;
               
           case QUAKE:
               break ;
               
          }
    }
*/
/********************** end of function ****************************/ 
 
/*
** Routine addsmooth
** addsmooth assembles the pressure smoothing vector.
*/
/*  - disabled -
 addsmooth( dest , rhs , info , el_pt )
     real  *dest , *rhs ;
     ELEMENT_INFO  *info ;
     ELEMENT_DATA  *el_pt ;

     {
      COUNTER      i    ;
      int          k    ;


           for ( i = 0 ; i < info->nen ; ++i )
               {
                k = el_pt->ien[i] - 1 ;
                *(dest + k) += rhs[i] ;
               }
     }
*/
/********************** end of function ****************************/ 
 
/*
** Routine apply_smooth
** apply_smooth modifies stress by applying pressure smoothing.
**/
/*  - disabled -
apply_smooth(grp_ptr)
GROUP   *grp_ptr ;
     {
      int     numel , nint , nel , j , ipt , nsd , nsdp1 ,
              ndof , nen , str_size , save_shape ;
      real  w_det , oldpress , pressure ;
      real  *sh , *shape_ptr , *det_ptr , *stress ;

      ELEMENT_DATA   *el_pt ;


      nsd = fe_sys.nsd ;
      nsdp1 = nsd + 1 ;
      ndof = fe_sys.ndof ;
      if( ndof == 1 ) return ;
      numel = (grp_ptr->el_info)->numel ;
      nen = (grp_ptr->el_info)->nen ;
      str_size = (grp_ptr->el_info)->stress_size ;
      save_shape = fe_sys.save_shape;
      
      switch((grp_ptr->el_info)->type)
          {
           case TET:
           case BILIN:
           case SERENDIP:
               for ( nel = 0 ; nel < numel ; nel++)
                   {
                    el_pt = grp_ptr->el_data + nel ;
                    nint = el_pt->nint ;

                    if( !save_shape )
                       {
                        shape_ptr = sh_temp ;
                        det_ptr = det_temp ;
                        shape( el_pt , grp_ptr->el_info , shape_ptr , det_ptr ,
                              (grp_ptr->el_info)->type ,
                              (grp_ptr->el_info)->nen , nint , 1 ) ;
                       }
                    else
                       {
                        shape_ptr = el_pt -> shape ;
                        det_ptr = el_pt -> deter ;
                       }

                    for (ipt = 0 ; ipt < nint ; ipt++)
                       {
                        pressure = ZERO ;
                        sh = shape_ptr + ipt*nen*nsdp1 ;
                        for(j=0;j<nen;j++)
                           {
                            pressure += SH(nsd,j) *
                            (*(global.pnode + el_pt->ien[j] - 1)) ;
                           }
                        stress = (el_pt->stress) + ipt*str_size ;
                        switch( ndof )
                          {
                           case 2:
                              oldpress = (*stress + *(stress + 1)) / TWO ;
                              *stress += (pressure - oldpress) / TWO ;
                              *(stress + 1) +=
                                         (pressure - oldpress) / TWO ;
                              break ;

                           case 3:
                              oldpress = (*stress + *(stress + 1)
                                          + *(stress + 2)) / THREE ;
                              *stress += (pressure - oldpress) / THREE ;
                              *(stress + 1) +=
                                         (pressure - oldpress) / THREE ;
                              *(stress + 2) +=
                                         (pressure - oldpress) / THREE ;
                              break ;
                          }
                       }  

                   }
               break ;
               
           case QUAKE:
               break ;
               
          }
    }
*/
/********************** end of function ****************************/ 
