/*
***                           File utility.c                      ***
***                        GeoFEST version 4.2                    ***
*** Copyright (c) 2002, California Institute of Technology        ***
*** U.S.Sponsorship under NASA Contract NAS7-1260 is acknowledged ***
***
*** This file contains miscellaneous utility routines used throughout
*** the finite element program: 
***
***          == move_real ==
***          == clear_real ==
***          == dot_real ==
***          == vadd ==
***          == vouter ==
***/

/*
***GeoFEST version 4.2***
G. A. Lyzenga -- modification to correctly relax hydrostatic stresses 6/15/00
J. W. Parker -- 3D support and deleted hypercube branches 9/16/98
G. A. Lyzenga  -- last modified on 3/03/95
Modification history:
2.1: fixes bug in boundary velocities and changes
     the way stresses are output.
2.2: implements split node faults
2.3: adds pass-thru comments
3.0: 3-D tetrahedra supported, hypercube branches deleted
3.1: hydrostatic relaxation changes
4.0: name change from visco to GeoFEST; correct implementation of 2-D out-of-
     plane stresses as well as hydrostatic relaxation; culling of remaining 
     vestiges of hypercube stuff; one code without recompilation for both
     2-D and 3-D domains
4.1: Introduction of trilinear hexahedral bricks and PCG iterative solver option
4.2: Bug fixes; standard source prologues added (June 02)
$Log: utility.c,v $
Revision 1.1.1.1  2002/07/12 19:05:14  jwp
Commiting GeoFEST to CVS at version 4.2

*/

#include <stdio.h>
#include "mydefs.h"
#include "errcodes.h"
#define EXTERN extern
#include "finel.h"


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   move_real   ***************
*/

/*  INTERFACE:   */

      move_real(
                real   *from , /* source array */
                real   *to ,   /* destination array */
                int    n       /* number of entries */
               )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine move_real
** move_real copies data to a new location.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      char CVS_Id[] = "@(#) CVS id $Id: utility.c,v 1.1.1.1 2002/07/12 19:05:14 jwp Exp $, Revision: " REVNO;
      real *last = to + n ;

      while( to < last ) *to++ = *from++ ;
     }
/************************** end of move_real *********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   clear_real   ***************
*/

/*  INTERFACE:   */

     clear_real(
                real   *array , /* array to be zapped */
                int    n       /* number of entries */
               )

/*  RETURN VALUE:  - none - */
/*  DESCRIPTION:   */
/*
** Routine clear_real
** clear_real nulls a data area.
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

     {
      real *last = array + n ;
      while(array < last ) *array++ = ZERO ;
     }
/************************** end of clear_real ********************************/ 


/*
/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   dot_real   ***************
*/

/*  INTERFACE:   */

  real dot_real(
                real   *vect_1 , /* multiplicand array */
                real   *vect_2 , /* multiplicand array */
                int    n       /* number of entries */
               )

/*  RETURN VALUE:  value of the scalar product */
/*  DESCRIPTION:   */
/*
** Function dot_real
** dot_real forms dot product of two vectors. 
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

      {
       int   i     ;
       real  sum    ;

       sum = ZERO ;
       for(i=0 ; i<n ; i++)
       {
             sum += (*(vect_1+i) * *(vect_2+i)) ;
       }

       return(sum) ;
      }
/************************** end of dot_real **********************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   vadd   ***************
*/

/*  INTERFACE:   */

           vadd(
                real   *v1 , /* first array */
                real   mult , /* scalar multiplier for second array */
                real   *v2 , /* second array */
                real   *dest , /* destination array */
                int    neq       /* number of entries */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine vadd
** calculates the linear combination of vectors
**    dest = v1 + mult*v2
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

  {
   int i ;
   
   for (i=0;i<neq;i++)
      *(dest+i) = *(v1+i) + mult * *(v2+i) ;
  }
  
/************************** end of vadd **************************************/ 


/*---------------------------------------------------------------------------*/
/*   BOP   */
/*
   *************   ROUTINE:   vouter   ***************
*/

/*  INTERFACE:   */

         vouter(
                real   *v1 , /* first array */
                real   *v2 , /* second array */
                real   *dest , /* destination array */
                int    neq       /* number of entries */
               )

/*  RETURN VALUE:  -none- */
/*  DESCRIPTION:   */
/*
** Routine vouter
** calculates the outer product of two vectors
** actually a misnomer -- it's really the element-by-element product
**/
/*   EOP   */
/*---------------------------------------------------------------------------*/

  {
   int i ;
   
   for (i=0;i<neq;i++)
      *(dest+i) = *(v1+i) * *(v2+i) ;
  }
  
/************************** end of vouter ************************************/ 
